#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
BB3 with OPT-175B Base.

SDM: Search Decision Model
MDM: Memory Decision Model
SGM: Search Query Generator Model
MGM: Memory Generator Model
MKM: Memory Knowledge Model
CKM: Contextual Knowledge Model
SKM: Search Knowledge Model
MRM: Memory Dialogue Response Model
CRM: Contextual Dialogue Response Model
SRM: Search Dialogue Response Model
"""
import copy
import random
from typing import List, Optional, Dict, Any, Union, Tuple
from parlai.agents.rag.retrievers import Document

from parlai.core.agents import Agent, create_agent
from parlai.core.message import Message
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from projects.bb3.agents.module import Module
from projects.bb3.agents.r2c2_bb3_agent import BlenderBot3Agent as R2C2Agent
from projects.bb3.agents.opt_api_agent import BB3OPTAgent
from projects.bb3.agents.search_agent import SearchAgent
from projects.bb3.agents.utils import (
    Decision,
    APIUtils,
    is_opener,
    DisplayUtils,
    set_failed_reply,
)
import projects.bb3.prompts as PROMPT


BB3_TO_OPT_KEYS = {
    'num_shots': 'num_shots',
    'opt_server': 'server',
    'include_prompt': 'include_prompt',
    'all_vanilla_prompt': 'all_vanilla_prompt',
    'metaseq_max_retry_api': 'max_retry_api',
    'metaseq_server_timeout': 'server_timeout',
    'memory_decision_use_memories': 'memory_decision_use_memories',
}

BB3_OPT_GEN_KEYS = [
    'prompt',
    'raw_prompt',
    'inference',
    'skip_generation',
    'beam_size',
    'beam_min_length',
    'beam_max_length',
    'topp',
    'temperature',
    'server',
    'omega_bound',
    'lambda_decay',
    'max_retry_api',
    'server_timeout',
    'alpha_presence',
    'alpha_frequency',
    'alpha_presence_src',
    'alpha_frequency_src',
]


class BlenderBot3Agent(R2C2Agent):
    """
    OPT BB3 Agent.
    """

    MODEL_TYPE: str = 'OPT'

    @classmethod
    def get_additional_agent_args(cls) -> ParlaiParser:
        parser = R2C2Agent.get_additional_agent_args()
        return BB3OPTAgent.add_cmdline_args(parser)

    def _apply_context_blocking_patches(self):
        """
        Don't do anything here.
        """
        pass

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Command line args for OPT BB3.
        """
        cls.add_additional_subagent_args(parser)
        parser = R2C2Agent.add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group('OPT Agent Group')
        group.add_argument(
            '--opt-server',
            type=str,
            help='which server to connect to for hosted OPT model',
            default=None,
        )
        group.add_argument(
            '--raw-search-server',
            type=str,
            help='specify a search server address.',
            default=None,
        )
        group.add_argument(
            '--num-shots',
            type=int,
            default=None,
            help='how many k-shot examples to put in the prompt. Default <0 is all',
        )
        group.add_argument(
            '--include-prompt',
            type='bool',
            default=None,
            help='Whether to include prompt',
        )
        group.add_argument(
            '--all-vanilla-prompt',
            type='bool',
            default=None,
            help='If True, and --include-prompt, then all prompts are just the vanilla, "a conversation", prompts.',
        )
        group.add_argument(
            '--knowledge-chunk-size',
            type=int,
            default=100,
            help='Chunk size, in words, of knowledge to keep',
        )
        group.add_argument('--debug-bb3', type='bool', default=False, dest='debug_bb3')
        group.add_argument(
            '--max-prompt-len',
            type=int,
            default=PROMPT.MAX_PROMPT_LEN,
            help='Longest sequence to send to API',
        )
        group.add_argument(
            '--self-prefix',
            type=str,
            default=PROMPT.SELF_PREFIX,
            help='Token for prefixing self responses',
        )
        group.add_argument(
            '--self-memory-prefix',
            type=str,
            default=PROMPT.SELF_MEMORY_PREFIX,
            help='Token for prefixing self memories',
        )
        group.add_argument(
            '--partner-prefix',
            type=str,
            default=PROMPT.PARTNER_PREFIX,
            help='Token for prefixing self responses',
        )
        group.add_argument(
            '--partner-memory-prefix',
            type=str,
            default=PROMPT.PARTNER_MEMORY_PREFIX,
            help='Token for prefixing self memories',
        )
        parser.add_argument(
            '--metaseq-max-retry-api',
            default=-1,
            type=int,
            help='Number of times to retry on API request failures (< 0 for unlimited retry).',
        )
        parser.add_argument('--metaseq-server-timeout', default=20.0, type=float)
        return parser

    def __init__(self, opt, shared=None):
        opt['model_file'] = 'zoo:'
        super().__init__(opt, shared)
        self.self_prefix = opt.get('self_prefix', PROMPT.SELF_PREFIX)
        self.partner_prefix = opt.get('partner_prefix', PROMPT.PARTNER_PREFIX)
        self.self_memory_prefix = opt.get(
            'self_memory_prefix', PROMPT.SELF_MEMORY_PREFIX
        )
        self.partner_memory_prefix = opt.get(
            'partner_memory_prefix', PROMPT.PARTNER_MEMORY_PREFIX
        )
        self.memory_utils.self_prefix = self.self_prefix
        self.memory_utils.partner_prefix = self.partner_prefix
        self.memory_utils.self_memory_prefix = self.self_memory_prefix
        self.memory_utils.partner_memory_prefix = self.partner_memory_prefix
        # Always init search agent
        if not shared:
            agent_opts = self.opts[Module.SEARCH_KNOWLEDGE]
            top_agent = self._init_top_agent(agent_opts)
            # two mappings; one for code access, one for debug access.
            self.agents[Module.SEARCH_KNOWLEDGE.agent_name()] = top_agent
            self.agents[Module.SEARCH_KNOWLEDGE] = top_agent
        else:
            top_agent = shared['top_agent'].clone()
            self.agents[Module.SEARCH_KNOWLEDGE.agent_name()] = top_agent
            self.agents[Module.SEARCH_KNOWLEDGE] = top_agent

        # continue
        self.max_prompt_len = opt.get('max_prompt_len', PROMPT.MAX_PROMPT_LEN)
        self.search_agent = SearchAgent(
            {
                'server': self.opt.get('search_server', 'default'),
                'raw_server': self.opt.get('raw_search_server', None),
                'intra_doc_delimiter': ' ',
                'n_docs': 5,
                'search_server_timeout': opt.get('search_server_timeout', 0),
            }
        )
        self.vanilla = all(
            decision is Decision.NEVER
            for decision in [
                self.contextual_knowledge_decision,
                self.search_decision,
                self.memory_decision,
            ]
        )
        self.batch_agents = {m: self.agents[m].clone() for m in Module}
        top_batch_agent = self.batch_agents[Module.SEARCH_KNOWLEDGE]
        for key in BB3_OPT_GEN_KEYS:
            for agent in self.batch_agents.values():
                if key not in agent.opt:
                    logging.debug(f"Overriding {key} to {top_batch_agent.opt[key]}")
                    agent.opt[key] = top_batch_agent.opt[key]

    def share(self):
        shared = super().share()
        shared['top_agent'] = self.agents[Module.SEARCH_KNOWLEDGE]
        return shared

    def _init_top_agent(self, opt: Opt) -> Agent:
        opt.pop('model_file', '')
        return create_agent(opt)

    def _construct_subagent_opts(self, opt: Opt):
        """
        Override to set prompts.
        """
        super()._construct_subagent_opts(opt)
        for m in Module:
            self.opts[m]['raw_prompt'] = "DUMMY"
            self.opts[m]['api_key'] = APIUtils.DEFAULT_KEY
            for self_k, agent_k in BB3_TO_OPT_KEYS.items():
                if opt[self_k] is not None:
                    old_val = self.opts[m].get(agent_k)
                    self.opts[m][agent_k] = opt[self_k]
                    logging.debug(
                        f"Overriding {agent_k} for {m} from {old_val} to {opt[self_k]}"
                    )

    def opening_token(self):
        return PROMPT.OPENING_PREFIX

    def get_mdm_observation(self, ag_obs: Message) -> Message:
        """
        Add memories to the memory decision observation.

        :param ag_obs:
            incoming mdm observation

        :return mdm_obs:
            return mdm observation with memories in the context.
        """
        self.agents[Module.MEMORY_DECISION].reset()

        original_text = ag_obs['text']
        self_memory_text, partner_memory_text = '', ''
        self_memories = [
            m.replace(f"{self.self_memory_prefix}: ", '')
            for m in self.memories
            if m.startswith(self.self_memory_prefix)
        ]
        if self_memories:
            self_memory_text = f"{PROMPT.MEMORY_KNOWLEDGE_PREFIX}: {self.self_memory_prefix}: {' '.join(self_memories)}\n"
        partner_memories = [
            m.replace(f"{self.partner_memory_prefix}: ", '')
            for m in self.memories
            if m.startswith(self.partner_memory_prefix)
        ]
        if partner_memories:
            partner_memory_text = f"{PROMPT.MEMORY_KNOWLEDGE_PREFIX}: {self.partner_memory_prefix}: {' '.join(partner_memories)}\n"

        new_text = f"{self_memory_text}{partner_memory_text}{original_text}"
        ag_obs.force_set('text', new_text)
        return ag_obs

    def get_orm_observation(
        self, observation: Message, opening_memories: Dict[str, int]
    ) -> Message:
        """
        Return the appropriate ORM observation.

        :param observation:
            raw observation
        :param opening_memories:
            memories seeded to the model

        :return observation:
            return the ORM observation
        """
        agent = self.agents[Module.OPENING_DIALOGUE]
        agent.reset()
        prefixed_memories = {}
        for mem, val in opening_memories.items():
            mem = self.memory_utils.maybe_add_memory_prefix(
                mem, 'partner', self.MODEL_TYPE
            )
            prefixed_memories[mem] = val

        new_obs = copy.deepcopy(observation)
        memories_to_use = self.memory_utils.get_available_memories(
            '',
            prefixed_memories,
            set(),
            dictionary=self.dictionary,
            **self._get_memory_heuristic_values(),
        )
        if not memories_to_use:
            # we need at least one memory to open with...
            memories_to_use = random.sample(list(prefixed_memories.keys()), 1)

        new_obs.force_set('text', self._check_and_limit_len('\n'.join(memories_to_use)))
        new_obs.force_set('memories', prefixed_memories)

        return agent.observe(new_obs)

    def get_opening_memories(
        self, memories: Optional[Union[List[str], Dict[str, int]]]
    ) -> Optional[Dict[str, int]]:
        """
        Get the opening memories, if applicable.

        This function is designed to handle legacy cases where memories are
        presented as a list, rather than a dict.

        :param memories:
            memories from the opening message

        :return opening_memories:
            return the set of true opening memories
        """
        opening_memories = None
        if memories:
            if isinstance(memories, dict):
                opening_memories = {mem: int(turns) for mem, turns in memories.items()}
            elif isinstance(memories, list):
                opening_memories = {}
                for mem in memories:
                    opening_memories = self.memory_utils.add_memory(
                        mem, opening_memories
                    )

        assert not opening_memories or isinstance(opening_memories, dict)
        return opening_memories

    def observe(self, observation: Message) -> Dict[Module, Message]:
        # handle passed memories as well
        observation = Message(observation)
        opening_memories = self.get_opening_memories(observation.pop('memories', None))
        observations = super().observe(observation)
        for m in Module.dialogue_modules():
            ag_obs = copy.deepcopy(observation)
            observations[m] = self.agents[m].observe(ag_obs)
        if is_opener(observation['text'], opening_memories):
            assert opening_memories
            orm_obs = self.get_orm_observation(observation, opening_memories)
            self.memories = orm_obs['memories']
            observations[Module.OPENING_DIALOGUE] = orm_obs
        else:
            observations[Module.OPENING_DIALOGUE] = Message({})

        if self.opt['debug_bb3']:
            DisplayUtils.display_observations(observations)
        return observations

    def _get_subagent_opt(
        self,
        filename: str,
        specific_override_args: Dict[str, Any],
        general_override_args: Dict[str, Any],
    ) -> Opt:
        # Remove the prefix for the model for the specific override args.
        specific_override_args = {
            '_'.join(k.split('_')[1:]): v for k, v in specific_override_args.items()
        }

        return {**general_override_args, **specific_override_args, 'override': {}}

    def get_search_results(
        self, batch_reply_sgm: List[Message], search_indices: List[int]
    ) -> Tuple[List[List[str]], List[List[Document]]]:
        """
        Retrieve search results; return documents as well.

        :param batch_reply_sgm:
            batch reply with the search queries
        :param search_indices:
            indices into batch in which we search.

        :return results, docs:
            return the search results in raw string form
            return the top documents as Documents
        """
        search_queries: List[str] = [
            o['text'] if i in search_indices else ''
            for i, o in enumerate(batch_reply_sgm)
        ]
        search_results: List[List[str]] = []
        top_docs: List[List[Document]] = []
        for q in search_queries:
            if q:
                search_results.append(
                    self.search_agent.respond({'text': q}).split('\n')
                )
                top_docs.append(self.search_agent.top_docs)
                # clear the top_docs on search agent side
                self.search_agent.top_docs = []
            else:
                search_results.append([])
                top_docs.append([])

        for i, results_i in enumerate(search_results):
            # chunk docs; save only first 500 chars of each document
            for j, d in enumerate(results_i):
                results_i[j] = ' '.join(
                    d.split(' ')[: self.opt['knowledge_chunk_size']]
                )
            print_results = '\t\n\n'.join(
                [
                    f'Document {j}:' + ' '.join(r.split()[:50])
                    for j, r in enumerate(results_i)
                ]
            )
            logging.debug(f"Search Results (50 toks each) for {i}:\n {print_results}")

        return search_results, top_docs

    def _check_and_limit_len(self, text: str) -> str:
        # check knowledge prompt len to make sure we don't overdo it.
        while len(self.dictionary.txt2vec(text)) >= self.max_prompt_len:
            text = '\n'.join(text.split('\n')[1:])
        return text

    def batch_act_knowledge(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        search_indices: List[int],
        memory_indices: List[int],
        contextual_indices: List[int],
        batch_agents: Dict[Module, Agent],
        available_memory: List[List[str]],
        search_results: List[List[str]],
        top_docs: List[List[Document]],
        top_memories: List[List[str]],
    ) -> List[Message]:
        """
        Add prompts to knowledge observations before calling batch act.

        :param observations:
            raw observations
        :param search_indices:
            indices in which we use search knowledge
        :param memory_indices:
            indices in which we use memories
        :param contextual_indices:
            indices in which we use an entity from the context
        :param batch_agents:
            mapping from module to agent with which to call batch act
        :param available_memory:
            memories per batch example
        :param search_results:
            search results per batch example
        :param top_docs:
            top docs corresponding to the search results
        :param top_memories:
            retrieved memories.

        :return batch_reply_knowledge:
            return batch reply from knowledge modules.
        """
        for i, all_obs in enumerate(observations):
            for module in Module:
                obs = all_obs[module]
                if module is Module.MEMORY_KNOWLEDGE and i in memory_indices:
                    memories = '\n'.join(available_memory[i])
                    new_prompt = self._check_and_limit_len(
                        obs['prompt'].replace(module.opt_pre_context_tok(), memories)
                    )
                    obs.force_set('prompt', new_prompt)
                elif module is Module.SEARCH_KNOWLEDGE and i in search_indices:
                    results = (
                        f"{PROMPT.EXTERNAL_KNOWLEDGE_PREFIX}: "
                        + f'\n{PROMPT.EXTERNAL_KNOWLEDGE_PREFIX}: '.join(
                            [s.replace('\n', '') for s in search_results[i]]
                        )
                    )
                    new_prompt = self._check_and_limit_len(
                        obs['prompt'].replace(module.opt_pre_context_tok(), results)
                    )
                    obs.force_set('prompt', new_prompt)

        batch_reply_knowledge = super().batch_act_knowledge(
            observations,
            search_indices,
            memory_indices,
            contextual_indices,
            batch_agents,
            top_docs=top_docs,
            top_memories=top_memories,
        )
        # re-assign memories
        for i, reply in enumerate(batch_reply_knowledge):
            if i in memory_indices:
                text = reply[Module.MEMORY_KNOWLEDGE.message_name()].strip()
                do_split = True
                if any(
                    text.startswith(p)
                    for p in [self.partner_memory_prefix, self.self_memory_prefix]
                ):
                    # no need to split the memory to find it
                    do_split = False
                true_memory = ''
                for memory in available_memory[i]:
                    raw_mem = (
                        self.memory_utils.split_prefix_memory(memory)[-1]
                        if do_split
                        else memory
                    )
                    if raw_mem == text:
                        true_memory = memory
                        break
                if true_memory:
                    reply.force_set(Module.MEMORY_KNOWLEDGE.message_name(), true_memory)
                else:
                    logging.debug(
                        f"Memory {text} is invalid, given {available_memory[i]}"
                    )
                    reply.pop(Module.MEMORY_KNOWLEDGE.message_name(), None)
                    reply.pop(f"{Module.MEMORY_KNOWLEDGE.message_name()}_docs", None)

        return batch_reply_knowledge

    def detect_and_handle_failures(
        self, module: Module, messages: List[Message], batch_reply_final: List[Message]
    ) -> List[int]:
        """
        Return indices of messages that have failures.

        :param messages:
            candidate list of Messages
        :param module:
            module in which failure may have occurred
        :param batch_reply_final:
            final batch replies

        :return indices:
            return indices of messages with failures
        """
        failed: List[int] = []
        for i, message in enumerate(messages):
            if APIUtils.is_request_failed_response(message):
                batch_reply_final[i]['failures'].extend(message['failures'])
                batch_reply_final[i][f"{module.message_name()}_failures"] = message[
                    'failures'
                ]
                failed.append(i)
        return failed

    def detect_and_handle_final_failures(
        self, messages: List[Message]
    ) -> List[Message]:
        """
        Handle failures of final dialogue replies.

        We get here if the whole pipeline collapses.
        """
        for reply in messages:
            if APIUtils.is_request_failed_response(reply):
                set_failed_reply(reply)
        return messages

    def batch_act_dialogue(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        batch_reply_knowledge: List[Message],
    ) -> List[Message]:
        """
        Batch act for dialogue!

        If we have failures of knowledge conditioning, fall back to vanilla dialogue.

        :param observations:
            raw observations per module
        :param batch_reply_knowledge:
            list of knowledge replies

        :return batch_act_dialogue:
            return the dialogue replies!
        """
        # construct dialogue obs with knowledge conditioning
        dialogue_obs, dialogue_replies = [], []
        for i, reply in enumerate(batch_reply_knowledge):
            for k_module in Module.knowledge_modules():
                if not reply.get(
                    k_module.message_name()
                ) or APIUtils.METASEQ_FAIL_MESSAGE_TEXT in reply.get(
                    k_module.message_name(), ''
                ):
                    continue
                logging.debug(
                    f'{k_module.message_name()}: {reply[k_module.message_name()]}'
                )
                d_module = None
                if k_module is Module.CONTEXTUAL_KNOWLEDGE:
                    d_module = Module.CONTEXTUAL_DIALOGUE
                elif k_module is Module.MEMORY_KNOWLEDGE:
                    d_module = Module.MEMORY_DIALOGUE
                else:
                    assert k_module is Module.SEARCH_KNOWLEDGE
                    d_module = Module.SEARCH_DIALOGUE
                d_obs = Message(observations[i][d_module])
                prompt = d_obs['prompt'].replace(
                    d_module.opt_post_context_tok(),
                    f"{d_module.opt_dialogue_knowledge_prefix()}{reply[k_module.message_name()]}",
                )
                d_obs.force_set('prompt', prompt)
                dialogue_obs.append((i, d_module, d_obs))

        # check if we have failures; if so, resort to vanilla dialogue
        all_inds = set([o[0] for o in dialogue_obs])
        for i in range(len(observations)):
            if i not in all_inds:
                dialogue_obs.append(
                    (
                        i,
                        Module.VANILLA_DIALOGUE,
                        Message(observations[i][Module.VANILLA_DIALOGUE]),
                    )
                )

        if self.opt['debug_bb3']:
            DisplayUtils.display_observations({o[1]: o[2] for o in dialogue_obs})

        # check knowledge conditioning; combine observations if so
        if self.knowledge_conditioning == 'separate' or len(dialogue_obs) == 1:
            dialogue_replies = self.batch_agents[Module.SEARCH_DIALOGUE].batch_act(
                [o[-1] for o in dialogue_obs]
            )
        elif self.knowledge_conditioning == 'combined':
            logging.debug('combining all knowledge')
            combined_obs: List[Message] = []
            for obs_i in range(len(observations)):
                extra_knowledge = []
                dialogue_obs_i = [d for d in dialogue_obs if d[0] == obs_i]
                primary_obs = dialogue_obs_i[0]
                for (_, _, obs) in dialogue_obs_i[1:]:
                    extra_knowledge.append(obs['prompt'].split('\n')[-2])
                old_prompt = primary_obs[-1]['prompt'].split('\n')
                primary_obs[-1].force_set(
                    'prompt',
                    '\n'.join(old_prompt[:-1] + extra_knowledge + old_prompt[-1:]),
                )
                combined_obs.append(primary_obs)
            batch_agent = self.batch_agents[self.combined_dialogue_module]
            dialogue_replies = batch_agent.batch_act([o[-1] for o in combined_obs])
            dialogue_obs = combined_obs
        else:
            raise NotImplementedError('Both is not implemented')

        batch_reply_dialogue = [Message({}) for _ in range(len(observations))]

        # collate dialogue replies
        for (i, d_module, _), reply in zip(dialogue_obs, dialogue_replies):
            if APIUtils.is_request_failed_response(reply):
                batch_reply_dialogue[i] = reply
                continue
            if self.knowledge_conditioning == 'combined':
                logging.debug(f'combined dialogue reply: {reply}')
                assert 'text' not in batch_reply_dialogue[i]
                batch_reply_dialogue[i]['text'] = reply['text']
                batch_reply_dialogue[i]['score'] = reply['logprobs']
                batch_reply_dialogue[i]['best_module'] = 'Combined Response'
            else:
                logging.debug(f'{d_module.message_name()}: {reply}')
            batch_reply_dialogue[i][d_module.message_name()] = reply['text']
            batch_reply_dialogue[i][f"{d_module.message_name()}_score"] = reply[
                'logprobs'
            ]

        # if separate, make sure we set text to highest scoring object
        if self.knowledge_conditioning == 'separate':
            for reply in batch_reply_dialogue:
                if APIUtils.is_request_failed_response(reply):
                    continue
                scores = []
                for module in Module.dialogue_modules():
                    scores.append(
                        (
                            module,
                            reply.get(f"{module.message_name()}_score", float('inf')),
                        )
                    )
                best_module = min(scores, key=lambda x: x[-1])[0]
                reply['text'] = reply[best_module.message_name()]
                reply['score'] = reply[f"{best_module.message_name()}_score"]
                reply['best_module'] = best_module

        return batch_reply_dialogue

    def batch_act_decision(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        module: Module,
        batch_reply_final: List[Message],
    ) -> Tuple[List[Message], List[int]]:
        """
        Override super class to handle failures.
        """
        agent = self.batch_agents[module]
        batch_reply, indices = super().batch_act_decision(observations, module, agent)
        failed = self.detect_and_handle_failures(module, batch_reply, batch_reply_final)
        indices = [i for i in indices if i not in failed]
        return batch_reply, indices

    def get_opening(
        self, observations: List[Dict[Union[str, Module], Message]]
    ) -> List[Message]:
        """
        Get the opening message.

        :param observations:
            raw observations

        :return batch_act:
            return batch act from the opening dialogue agent.
        """
        module = Module.OPENING_DIALOGUE
        batch_act = [Message({'text': self.self_prefix})] * len(observations)

        def _failed_messages(replies):
            return any(
                p in o['text']
                for p in [self.self_prefix, self.partner_prefix]
                for o in replies
            )

        retries = 0
        opening_obs = [o[module] for o in observations]
        while _failed_messages(batch_act) and retries < 3:
            batch_act = self.batch_agents[Module.OPENING_DIALOGUE].batch_act(
                opening_obs
            )
            self.batch_agents[Module.OPENING_DIALOGUE].reset()
            retries += 1
            n_mems = [min(1, len(obs['memories']) // 3) for obs in opening_obs]
            for i, o in enumerate(opening_obs):
                mem_indices = random.sample(range(len(o['memories'])), n_mems[i])
                o.force_set(
                    'memories',
                    {
                        m: v
                        for i, (m, v) in enumerate(o['memories'].items())
                        if i in mem_indices
                    },
                )
        if _failed_messages(batch_act):
            for reply in batch_act:
                text = reply.pop('text')
                for p in [
                    self.self_memory_prefix,
                    self.partner_memory_prefix,
                    self.self_prefix,
                    self.partner_prefix,
                ]:
                    text = text.replace(f"{p}:", '').replace(p, '')
                reply['text'] = text

        return batch_act

    def batch_act_simple(
        self,
        observations: List[Dict[Union[str, Module], Message]],
        module: Module,
    ) -> List[Message]:
        """
        Return either vanilla batch act or opening batch act.

        :param observations:
            list of observations
        :param module:
            module to perform simple batch act with

        :return batch_act:
            return batch act from appropriate agent module.
        """
        batch_agent = self.batch_agents[module]
        if module is Module.OPENING_DIALOGUE:
            batch_act = self.get_opening(observations)
        else:
            batch_act = batch_agent.batch_act([o[module] for o in observations])

        batch_reply_final = [Message({}) for _ in range(len(observations))]
        failed = self.detect_and_handle_failures(module, batch_act, batch_reply_final)

        for i, reply in enumerate(batch_act):
            if i in failed:
                set_failed_reply(reply)
            else:
                reply[module.message_name()] = reply['text']
                reply[f"{module.message_name()}_score"] = reply.get(
                    'logprobs', float('inf')
                )
        batch_reply_final = self.collate_batch_acts(
            batch_reply_mdm=[Message({})] * len(observations),
            batch_reply_sdm=[Message({})] * len(observations),
            batch_reply_mgm_partner=[Message({})] * len(observations),
            batch_reply_mgm_self=[Message({})] * len(observations),
            batch_reply_sgm=[Message({})] * len(observations),
            batch_reply_knowledge=[Message({})] * len(observations),
            batch_reply_dialogue=batch_act,
            available_memory=[o['raw']['memories'] for o in observations],
        )
        return self.detect_and_handle_final_failures(batch_reply_final)

    def batch_act(
        self, observations: List[Dict[Union[str, Module], Message]]
    ) -> List[Message]:
        """
        Override batch act for OPT BB3.
        """
        if self.vanilla:
            return self.batch_act_simple(observations, Module.VANILLA_DIALOGUE)

        if all(
            is_opener(o['raw']['text'], o[Module.OPENING_DIALOGUE].get('memories'))
            for o in observations
        ):
            # We're opening!
            return self.batch_act_simple(observations, Module.OPENING_DIALOGUE)

        batch_reply_final = [
            Message({'id': 'BlenderBot3', 'failures': []})
            for _ in range(len(observations))
        ]
        # Step 1: determine whether we're searching or accessing memory
        all_memory: List[Dict[str, int]] = [o['raw']['memories'] for o in observations]
        available_memory = self.get_available_memories(observations)

        batch_reply_sdm, search_indices = self.batch_act_decision(
            observations,
            Module.SEARCH_DECISION,
            batch_reply_final,
        )
        batch_reply_mdm, memory_indices = self.batch_act_decision(
            observations,
            Module.MEMORY_DECISION,
            batch_reply_final,
        )

        memory_indices = [i for i in memory_indices if available_memory[i]]
        contextual_indices = [
            i
            for i in range(len(observations))
            if (
                self.contextual_knowledge_decision is Decision.ALWAYS
                or (
                    self.contextual_knowledge_decision is Decision.COMPUTE
                    and i not in search_indices + memory_indices
                )
            )
            and self.contextual_knowledge_decision is not Decision.NEVER
        ]
        all_inds = (
            set(contextual_indices)
            .union(set(search_indices))
            .union(set(memory_indices))
        )
        if not all(i in all_inds for i in range(len(observations))):
            # something went terribly, terribly wrong. Vanilla batch act.
            return self.batch_act_simple(observations, Module.VANILLA_DIALOGUE)
        # Step 2: Generate search queries, if necessary. Also generate memories for partner
        batch_reply_mgm_partner = self.batch_act_mgm(
            observations=observations, agent=self.batch_agents[Module.MEMORY_GENERATOR]
        )
        batch_reply_sgm = self.batch_act_sgm(
            observations, search_indices, self.batch_agents[Module.SEARCH_QUERY]
        )
        search_indices = [
            i
            for i in search_indices
            if i
            not in self.detect_and_handle_failures(
                Module.SEARCH_QUERY, batch_reply_sgm, batch_reply_final
            )
        ]

        search_results, top_docs = self.get_search_results(
            batch_reply_sgm, search_indices
        )
        invalid_search_results = [i for i in search_indices if not top_docs[i]]
        contextual_indices += invalid_search_results
        search_indices = [i for i in search_indices if i not in invalid_search_results]

        # step 3: memory/knowledge/contextual access.
        # for knowledge obs, we need to fill in context with external or persona knowledge
        batch_reply_knowledge = self.batch_act_knowledge(
            observations,
            search_indices,
            memory_indices,
            contextual_indices,
            {m: self.batch_agents[m] for m in Module if m.is_knowledge()},
            available_memory,
            search_results,
            top_docs,
            [
                [Document(docid='', text=memory, title='') for memory in memories]
                for memories in available_memory
            ],
        )

        # step 4: memory/knowledge/contextual dialogue
        batch_reply_dialogue = self.batch_act_dialogue(
            observations, batch_reply_knowledge
        )
        batch_reply_mgm_self = self.batch_act_mgm(
            self_messages=[
                self.agents[Module.MEMORY_GENERATOR].observe(r)
                for r in batch_reply_dialogue
            ],
            agent=self.batch_agents[Module.MEMORY_GENERATOR],
        )

        batch_reply_final = self.collate_batch_acts(
            batch_reply_sdm,
            batch_reply_mdm,
            batch_reply_sgm,
            batch_reply_mgm_self,
            batch_reply_mgm_partner,
            batch_reply_knowledge,
            batch_reply_dialogue,
            all_memory,
        )
        for i, reply in enumerate(batch_reply_final):
            reply.force_set('id', 'BlenderBot3')
            if i in invalid_search_results:
                reply.force_set('search_result_failure', True)
            if self.opt['debug_bb3']:
                DisplayUtils.display_act(reply)

        batch_reply_final = self.detect_and_handle_final_failures(batch_reply_final)

        return batch_reply_final

    def self_observe(self, self_message: Message):
        logging.debug(f"Self-observing message: {self_message}")
        if APIUtils.is_request_failed_response(self_message):
            return

        for m, agent in self.agents.items():
            if isinstance(m, Module):
                logging.debug(f'Self-observing for module {m}')
                agent.self_observe(self_message)
        if self.vanilla:
            return
        self.self_observe_memory(self_message)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.teachers import DialogTeacher
from parlai.utils.io import PathManager
from parlai.utils.data import DatatypeHelper
from .build import build
import os
import pandas as pd
from typing import Optional


class MetalWozTeacherBase(DialogTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            "--metalwoz-domains", nargs="+", help="Use only a subset of the domains"
        )
        return parser

    def _path(self, opt):
        fold = DatatypeHelper.fold(opt["datatype"])
        if fold == "train" or fold == "valid":
            folder = os.path.join(opt["datapath"], "metalwoz", "train")
        else:
            folder = os.path.join(opt["datapath"], "metalwoz", "test")
        return folder, fold

    def __init__(self, opt, shared=None):
        if shared is None:
            build(opt)
        folder, fold = self._path(opt)
        self.fold = fold
        opt["datafile"] = os.path.join(folder, fold)
        super().__init__(opt, shared)

    def load_data(self, datapath):
        folder, fold = os.path.split(datapath)
        with PathManager.open(os.path.join(folder, "tasks.txt")) as taskf:
            tasks_table = pd.read_json(taskf, lines=True)

        dfolder = os.path.join(folder, "dialogues")

        data = []

        for filename in PathManager.ls(dfolder):
            domain = filename.replace(".txt", "")
            if (
                self.opt["metalwoz_domains"]
                and domain not in self.opt["metalwoz_domains"]
            ):
                continue
            fullfn = os.path.join(dfolder, filename)
            with PathManager.open(fullfn) as dataf:
                lines = pd.read_json(dataf, lines=True)
                lines = lines.merge(tasks_table, on="task_id")
                data.append(lines.to_dict("records"))

        # Quick check to make sure we didn't fat-finger the spelling of some domain
        if self.opt["metalwoz_domains"]:
            assert len(data) == len(self.opt["metalwoz_domains"])

        if "test" in self.fold:
            flat = []
            for domain in data:
                flat.extend(domain)
            return flat

        return DatatypeHelper.split_subset_data_by_fold(self.fold, data, 0.8, 0.1, 0.1)


class SystemTeacher(MetalWozTeacherBase):
    def setup_data(self, datapath):
        data = self.load_data(datapath)
        for row in data:
            texts = [row["bot_role"]] + list(row["turns"])
            prompts, labels = texts[::2], texts[1::2]
            for i, (prompt, label) in enumerate(zip(prompts, labels)):
                yield {
                    "text": prompt,
                    "label": label,
                    "bot_role": row["bot_role"],
                    "bot_prompt": row["bot_prompt"],
                    "user_role": row["user_role"],
                    "user_prompt": row["user_prompt"],
                    "utterance_id": row["id"],
                    "domain": row["domain_x"],
                    "task_id": row["task_id"],
                }, i == 0


class UserSimulatorTeacher(MetalWozTeacherBase):
    def setup_data(self, datapath):
        data = self.load_data(datapath)
        for row in data:
            texts = list(row["turns"])
            prompts, labels = (
                [f"{row['user_role']}\n{texts[0]}"] + texts[2::2],
                texts[1::2],
            )
            for i, (prompt, label) in enumerate(zip(prompts, labels)):
                yield {
                    "text": prompt,
                    "label": label,
                    "bot_role": row["bot_role"],
                    "bot_prompt": row["bot_prompt"],
                    "user_role": row["user_role"],
                    "user_prompt": row["user_prompt"],
                    "utterance_id": row["id"],
                    "domain": row["domain_x"],
                    "task_id": row["task_id"],
                }, i == 0


class MetalWozTeacher(SystemTeacher):
    pass


class DefaultTeacher(MetalWozTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
MultiDoGo implementation for ParlAI.

NOTE: There is still missing data in the open source version of this; implementation is not complete. See https://github.com/awslabs/multi-domain-goal-oriented-dialogues-dataset/issues/1
"""

from typing import Optional
from parlai.core.params import ParlaiParser
import copy
import json
import os
from parlai.core.opt import Opt
from parlai.utils.data import DatatypeHelper
import parlai.core.tod.tod_core as tod
import parlai.core.tod.tod_agents as tod_agents
import parlai.tasks.multidogo.build as build_
from parlai.tasks.multidogo.build import get_processed_multidogo_folder
from parlai.tasks.multidogo.build import (
    DOMAINS,
    SENTENCE_INTENT,
    TURN_INTENT,
    TURN_AND_SENTENCE_INTENT,
)

INTENT_ANNOTATION_TYPES = [SENTENCE_INTENT, TURN_INTENT, TURN_AND_SENTENCE_INTENT]


class MultidogoParser(tod_agents.TodStructuredDataParser):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            "--multidogo-domains",
            nargs="+",
            default=DOMAINS,
            choices=DOMAINS,
            help="Uses last passed in configuration.",
        )
        parser.add_argument(
            "--intent-type",
            default=TURN_INTENT,
            choices=INTENT_ANNOTATION_TYPES,
            help="Sets the type of intent classification labels. Sentence annotations represented as a list with adjacent entries of the same type deduped.",
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        self.dpath = os.path.join(opt["datapath"], "multidogo")
        opt["datafile"] = self.fold
        build_.build(opt)
        super().__init__(opt, shared)

    def setup_episodes(self, fold):
        result = []
        domains = self.opt.get("multidogo_domains", DOMAINS)
        if type(domains) is str:
            domains = [domains]
        intent_type = self.opt.get("intent-type", TURN_INTENT)
        for _conv_id, domain, conversation in self._iterate_over_conversations(
            domains, intent_type
        ):
            if len(conversation) == 0 or not (
                all(["role" in turn for turn in conversation.values()])
            ):
                continue
            rounds = []
            prev_role = conversation["0"]["role"]
            if prev_role == "customer":
                user_utt = [conversation["0"]["text"]]
                api_call = conversation["0"].get("slots", {})
                api_resp = {}
                sys_utt = []
            else:
                user_utt = ["__SILENCE__"]
                api_call = {}
                api_resp = conversation["0"].get("slots", {})
                sys_utt = [conversation["0"]["text"]]
            all_calls = api_call
            api_call = {tod.STANDARD_API_NAME_SLOT: domain}
            for i in range(1, len(conversation)):
                turn = conversation[str(i)]
                if prev_role == "agent" and prev_role != turn["role"]:
                    rounds.append(
                        tod.TodStructuredRound(
                            user_utt="\n".join(user_utt),
                            api_call_machine=api_call,
                            api_resp_machine=api_resp,
                            sys_utt="\n".join(sys_utt),
                        )
                    )
                    user_utt = []
                    api_call = {tod.STANDARD_API_NAME_SLOT: domain}
                    api_resp = {}
                    sys_utt = []
                prev_role = turn["role"]
                slot = turn.get("slots", {})
                if prev_role == "customer":
                    user_utt.append(turn["text"])
                    api_call.update(slot)
                    all_calls.update(slot)
                else:
                    api_resp.update(slot)
                    sys_utt.append(turn["text"])

            rounds.append(
                tod.TodStructuredRound(
                    user_utt="".join(user_utt),
                    api_call_machine=api_call,
                    api_resp_machine=api_resp,
                    sys_utt="".join(sys_utt),
                )
            )
            goal_calls = copy.deepcopy(all_calls)
            goal_calls[tod.STANDARD_API_NAME_SLOT] = domain
            result.append(
                tod.TodStructuredEpisode(
                    domain=domain,
                    api_schemas_machine=[
                        {
                            tod.STANDARD_API_NAME_SLOT: domain,
                            tod.STANDARD_OPTIONAL_KEY: all_calls.keys(),
                        }
                    ],
                    goal_calls_machine=[goal_calls],
                    rounds=rounds,
                )
            )
        return result

    def _iterate_over_conversations(self, domains, intent):
        for domain in domains:
            data_folder = get_processed_multidogo_folder(
                self.dpath, domain, self.fold, intent
            )
            for filename in os.listdir(data_folder):
                if filename.endswith(".json"):
                    with open(data_folder + "/" + filename) as f:
                        data = json.load(f)
                        for conv_id, value in data.items():
                            yield conv_id, domain, value

    def get_id_task_prefix(self):
        return "Multidogo"


class SystemTeacher(MultidogoParser, tod_agents.TodSystemTeacher):
    pass


class UserSimulatorTeacher(MultidogoParser, tod_agents.TodUserSimulatorTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from projects.bb3.agents.module import Module
import random

#######################
# Important Constants #
#######################
PARTNER_PREFIX = 'Person 1'
SELF_PREFIX = 'Person 2'
QUERY_GEN_PREFIX = 'Query'
MEMORY_GEN_PREFIX = 'Memory'
MEMORY_KNOWLEDGE_PREFIX = 'Personal Fact'
CONTEXTUAL_KNOWLEDGE_PREFIX = 'Previous Topic'
SEARCH_KNOWLEDGE_PREFIX = 'Interesting Fact'
STYLE_PREFIX = 'Style'
OPENING_PREFIX = 'Opening'
CONTEXT_PREFIX = 'Context'

SELF_MEMORY_PREFIX = f"{SELF_PREFIX}'s Persona"
PARTNER_MEMORY_PREFIX = f"{PARTNER_PREFIX}'s Persona"
EXTERNAL_KNOWLEDGE_PREFIX = "External Knowledge"

SEARCH = 'search'
DO_NOT_SEARCH = 'do not search'
SEARCH_DECISION = "Search Decision"
MEMORY_DECISION = "Memory Decision"

ACCESS_MEMORY = "access memory"
DO_NOT_ACCESS_MEMORY = "do not access memory"

NO_MEMORY = "no persona"

PRE_CONTEXT_TOK = "__REPLACE_ME_PRE_CONTEXT__"
POST_CONTEXT_TOK = "__REPLACE_ME_POST_CONTEXT__"

MAX_PROMPT_LEN = 1912

###################
# LIGHT CONSTANTS #
###################
LIGHT_SETTING_NAME = "Setting"
LIGHT_SETTING_DESC = "Setting Description"
LIGHT_PARTNER_NAME = f"{PARTNER_PREFIX} is"
LIGHT_SELF_NAME = f"{SELF_PREFIX} is"

SEARCH_DECISION_STRING = f"""
{PARTNER_PREFIX}: Who directed 2001: A Space Odyssey?
{SEARCH_DECISION}: {SEARCH}

{PARTNER_PREFIX}: Do you like italian food as well?
{SEARCH_DECISION}: {DO_NOT_SEARCH}

{PARTNER_PREFIX}: On which channel will the FIFA world cup be broadcasted?
{SEARCH_DECISION}: {SEARCH}

{PARTNER_PREFIX}: oh gosh, I'm so sad now. That's terrible
{SEARCH_DECISION}: {DO_NOT_SEARCH}

{PARTNER_PREFIX}: To whom was John B. Kroc married?
{SEARCH_DECISION}: {SEARCH}

{PARTNER_PREFIX}: have you ever adopted an animal from animal shelters? I did, i adopted a chinchilla and a cat
{SEARCH_DECISION}: {DO_NOT_SEARCH}

{PARTNER_PREFIX}: Which is the largest of the Japanese Volcano Islands?
{SEARCH_DECISION}: {SEARCH}

{PARTNER_PREFIX}: how cool! are you going to go?
{SEARCH_DECISION}: {DO_NOT_SEARCH}

{PARTNER_PREFIX}: I did not know he did media other than just his children stories books. Did Dr. Seuss do anything else other than paintings and books, such as movies or scripts?
{SEARCH_DECISION}: {SEARCH}"""


MEMORY_DECISION_STRING = f"""
{PARTNER_PREFIX}: I started reading a new novel series. It's very much like agatha christie. I think I found something to keep me going.
{MEMORY_DECISION}: {ACCESS_MEMORY}

{PARTNER_PREFIX}: No. He was an author, but has been retired for a couple of years now.
{MEMORY_DECISION}: {DO_NOT_ACCESS_MEMORY}

{PARTNER_PREFIX}: No plans right now, but I'd love to go back to ireland. I stayed in dublin last time I was there. Would love to visit the other side of the island. Maybe around galway. Are you irish? That's wht I went. Wanted to see the ancestral home.
{MEMORY_DECISION}: {ACCESS_MEMORY}

{PARTNER_PREFIX}: I'm very sorry for your loss.
{MEMORY_DECISION}: {DO_NOT_ACCESS_MEMORY}

{PARTNER_PREFIX}: It is beautiful there. If you get the chance its definitely worth the trip. How old are your children?
{MEMORY_DECISION}: {ACCESS_MEMORY}

{MEMORY_DECISION}: I think corn may be easier on the stomach, otherwise it's just too starchy!
{MEMORY_DECISION}: {DO_NOT_ACCESS_MEMORY}

{PARTNER_PREFIX}: What kind of job are you looking for?
{MEMORY_DECISION}: {ACCESS_MEMORY}

{PARTNER_PREFIX}: Yeah, I'll send you a link. I saw an ad for it earlier this week. Unfortunately, I haven't been swimming much it's been really cold outside! I wish my gym had an indoor heated pool.
{MEMORY_DECISION}: {DO_NOT_ACCESS_MEMORY}

{PARTNER_PREFIX}: That's lovely. I have 1 dog and 1 cat. And you?
{MEMORY_DECISION}: {ACCESS_MEMORY}"""


SEARCH_QUERY_STRING = f"""
{PARTNER_PREFIX}: My favorite song is Rock On By David Essex. Most people don't even know this song, but it is amazing and everyone should hear it.
{SELF_PREFIX}: I'm not too familiar with David Essex's Rock On. Tell me something about the song.
{PARTNER_PREFIX}: It has a unique bass line. The lyrics speak to me. It is a classic rock song from 1973
{QUERY_GEN_PREFIX}: Rock On David Essex lyrics

{PARTNER_PREFIX}: My favorite athlete is nadal; he will be the greatest tennis player.
{SELF_PREFIX}: Tell about your hobbies in 2021!
{PARTNER_PREFIX}: I will continue to play tennis. I hope that I can hit the ball once a day.
{SELF_PREFIX}: Do you know about Rafael?
{PARTNER_PREFIX}: Rafa is my favourite tennis player
{QUERY_GEN_PREFIX}: interesting facts about rafael

{PARTNER_PREFIX}: My favorite item to buy is Chocolate bars. My favourite is Oh Henry. I can never resist buying a chocolate bar when I am at the grocery store.
{SELF_PREFIX}: Did you ever try the premium chocolates boxes for yourself or sending gifts to someone?
{PARTNER_PREFIX}: I have never seem them. Do they have a good variety of chocolate?
{SELF_PREFIX}: Yes they do. They have much more than boxed chocolates and bars. They have different kind of flavors like caramels, toffee bars etc
{PARTNER_PREFIX}: It sounds wonderful. I wonder if there are any nut-free varieties or low sugar options.
{QUERY_GEN_PREFIX}: Low sugar chocolate

{PARTNER_PREFIX}: My favorite author is john grisham, and i play paintball on the weekends. Have you ever been shot by a paintball?
{SELF_PREFIX}: No, but I could imagine it probably hurts!
{PARTNER_PREFIX}: yes it does, especially if the paintballs are frozen!
{SELF_PREFIX}: Frozen? I didn't even know you could freeze them! What would you compare the feeling to? I've heard a lot of people will sprain their ankles or strain a muscle while out playing!
{PARTNER_PREFIX}: i would equate it to a pretty bad wasp sting.  yeah i have definitely sprained an ankle out there.
{QUERY_GEN_PREFIX}: sprained ankle healing

{PARTNER_PREFIX}: My favorite actor is Tom Hanks. It's absolutely incredible how many hit movies Tom Hanks has put out. What is your favorite Tom Hanks movie of them all?
{QUERY_GEN_PREFIX}: Tom Hanks movies"""


MEMORY_GENERATOR_STRING = f"""
{SELF_PREFIX}: Hello, whats your favorite color? Mine is yellow.
{PARTNER_PREFIX}: Good, thank you! Tell me a secret about your life.
{SELF_PREFIX}: I believe in ghosts, and I am fascinated by them.
{PARTNER_PREFIX}: I have not once kissed a woman. I win! Lol
{SELF_PREFIX}: Oh yeah, well my childhood dream was to be an architect.
{PARTNER_PREFIX}: Nice! What career did you end up in?
{SELF_PREFIX}: I plan weddings now, what do you do?
{PARTNER_PREFIX}: Nice! I am in business. I surprised of where I am, really.
{MEMORY_GEN_PREFIX}: {PARTNER_PREFIX} works in the business sector.

{SELF_PREFIX}: The Target app is awesome! Is this what you use to shop?
{PARTNER_PREFIX}: No, I usually just go to the store and shop. There's a Target near my home.
{MEMORY_GEN_PREFIX}: {PARTNER_PREFIX} lives near a Target.

{SELF_PREFIX}: Have you found any luck finding a second job?
{PARTNER_PREFIX}: Hey, not yet. I'm just so tired all the time. Working nights is hard, and during the day I'm working on the details of trying to open my own restaurant, but that's hard to do when you're as broke as me, but it's my dream, so I'm going to work at it as much as it takes. I'll keep looking for a second job to save up the money. {SELF_PREFIX}'s business doesn't happen to be hiring right now are they?
{MEMORY_GEN_PREFIX}: {PARTNER_PREFIX} has a job, wants to open my own restaurant, works at night, wants to find a second job, and doesn't have a lot of money.

{PARTNER_PREFIX}: Hello. What are you up to today?
{SELF_PREFIX}: Same as everyday, just sitting at home listening to backstreet boys. You?
{PARTNER_PREFIX}: Love the band. I'm going rock climbing with my dog
{MEMORY_GEN_PREFIX}: {PARTNER_PREFIX} is rock climbing with their dog. {PARTNER_PREFIX} loves the Backstreet Boys.

{PARTNER_PREFIX}: I just took some great photographs outdoors.
{MEMORY_GEN_PREFIX}: {NO_MEMORY}"""


MEMORY_KNOWLEDGE_STRING = f"""
{PARTNER_MEMORY_PREFIX} I do not have any pets. I like all sports. I enjoy coaching volleyball and my volleyball team is my favorite team.
{PARTNER_MEMORY_PREFIX} I enjoy doing crosswords. I enjoy hiking.
{PARTNER_MEMORY_PREFIX} I speak french and english fluently.
{PARTNER_MEMORY_PREFIX} I've two border collies.
{SELF_MEMORY_PREFIX} I own my own art studio.
{SELF_MEMORY_PREFIX} I enjoy going on hikes.
{SELF_MEMORY_PREFIX} I speak French and French food is my favorite.
{SELF_MEMORY_PREFIX} I do not have cats because I am allergic to cats. I have two collies.
{SELF_MEMORY_PREFIX} I coach the girl s volley ball team.
{SELF_MEMORY_PREFIX} I am working on art projects. I am moving.
{SELF_MEMORY_PREFIX} I am moving my studio. I am an artist.
{SELF_MEMORY_PREFIX} I didn't study art in school. I've thought about doing school tours, participated in art-focused assemblies and the like, but not about teaching art full time.
{SELF_PREFIX}: Hi. How are you?
{PARTNER_PREFIX}: I am good. I just got home from work. You?
{SELF_PREFIX}: Good. Cleaning up my art studio. Where do you work?
{PARTNER_PREFIX}: I teach middle school. I relax on weekends by singing in a bar.
{SELF_PREFIX}: How fun. I go on hikes. Do you have pets?
{PARTNER_PREFIX}: No pets for me. I do like sports though. Football is my favorite.
{SELF_PREFIX}: Do you have a favorite team?
{MEMORY_KNOWLEDGE_PREFIX}: {PARTNER_PREFIX} coaches a girl's volley ball team.

{PARTNER_MEMORY_PREFIX} I work at the church food pantry.
{PARTNER_MEMORY_PREFIX} I have 5 children, 2 of them have died.
{PARTNER_MEMORY_PREFIX} I plant sunflowers on my children's graves.
{PARTNER_MEMORY_PREFIX} I love to read agatha christie.
{SELF_MEMORY_PREFIX} My mother was born in ireland.
{PARTNER_MEMORY_PREFIX} I like to travel.
{PARTNER_MEMORY_PREFIX} My father was an author.
{SELF_MEMORY_PREFIX} I went to Australia.
{SELF_MEMORY_PREFIX} I've been to Ireland. My mother is from Ireland.
{SELF_MEMORY_PREFIX} My father was an author.
{SELF_MEMORY_PREFIX} I love to read. I've read the black stallion.
{SELF_MEMORY_PREFIX} I grow the tallest sunflowers you have ever seen.
{PARTNER_MEMORY_PREFIX} I have never been to Ireland. I want to travel to Ireland soon.
{SELF_MEMORY_PREFIX} I have been to Dublin. I would like to visit Galway.
{PARTNER_MEMORY_PREFIX} My twins were born premature. I have twins.
{SELF_PREFIX}: Hello. I just got back from australia. How are you?
{PARTNER_PREFIX}: I am well. Just finished working at the church food pantry where o am. Most days.
{SELF_PREFIX}: I also like to go to ireland. It is where my mother was from.
{PARTNER_PREFIX}: I have 5 lovely children that are my world. 2 have died
{SELF_PREFIX}: I am sorry to hear that. Do you have fun hobbies?
{PARTNER_PREFIX}: The tallest sunflowers you have ever seen are over their graves
{SELF_PREFIX}: Wow that sounds wonderful! Did you plant them there?
{PARTNER_PREFIX}: I grow them as well as read agatha christie novels
{SELF_PREFIX}: My father liked those books too and he was a great famous author.
{PARTNER_PREFIX}: I love reading them. Do you like to read?
{MEMORY_KNOWLEDGE_PREFIX}: {SELF_PREFIX} has read the black stallion."""


CONTEXTUAL_KNOWLEDGE_STRING = f"""
{SELF_PREFIX}: I like to remodel homes, go hunting, and shoot a bow.
{SELF_PREFIX}: My favorite holiday is halloween.
{PARTNER_PREFIX}: Hi, how are you doing? I'm getting ready to do some cheetah chasing to stay in shape.
{SELF_PREFIX}: You must be very fast. Hunting is one of my favorite hobbies.
{PARTNER_PREFIX}: I am! For my hobby I like to do canning or some whittling.
{CONTEXTUAL_KNOWLEDGE_PREFIX}: homes

{SELF_PREFIX}: My mom is my best friend, I have four sisters, I believe that mermaids are real, and I love iced tea.
{PARTNER_PREFIX}: Hi, how are you doing today?
{SELF_PREFIX}: I am spending time with my 4 sisters, what are you up to?
{PARTNER_PREFIX}: Wow, four sisters. Just watching game of thrones.
{SELF_PREFIX}: That is a good show, I watch that while drinking iced tea.
{PARTNER_PREFIX}: I agree. What do you do for a living?
{CONTEXTUAL_KNOWLEDGE_PREFIX}: mermaids

{SELF_PREFIX}: I had a gig at local theater last night, I work as a stand up comedian, I come from a small town, my favorite drink is cuba libre, and I did a few small roles in tv series.
{PARTNER_PREFIX}: We all live in a yellow submarine, a yellow submarine. Morning!
{SELF_PREFIX}: Hi! That is a great line for my next stand up.
{PARTNER_PREFIX}: Lol. I am shy, anything to break the ice, and I am a beatles fan.
{CONTEXTUAL_KNOWLEDGE_PREFIX}: tv

{SELF_PREFIX}: I like horseback riding, my favorite food is cheese, I'm allergic to shellfish, I work at a non profit that helps children, and I love going to concerts and dancing hard.
{PARTNER_PREFIX}: Hello, I've to be quick because I work in a hospital
{CONTEXTUAL_KNOWLEDGE_PREFIX}: children

{SELF_PREFIX}: I enjoy fishing, I have a dog named bob, I live on an island, and I like to make boats on the weekends.
{PARTNER_PREFIX}: Hi, how are you today?
{SELF_PREFIX}: I am great just about to head out fishing one of my favorite pastimes
{PARTNER_PREFIX}: Sounds fun! I am going shoe shopping later.
{SELF_PREFIX}: On the island that I live on there are many places to shop
{PARTNER_PREFIX}: I would shop all day! What do you like to do in your spare time?
{CONTEXTUAL_KNOWLEDGE_PREFIX}: weekends"""


SEARCH_KNOWLEDGE_STRING = f"""
{EXTERNAL_KNOWLEDGE_PREFIX}: His Military Charity Work. This Is Us: Susan Kelechi Watson on How Beth Is Finding Her Identity Apart From Randall. Lego’s new Toy Story 4 tie-in sets unveiled. Sun, Feb 17 12:30 AM EST on AMC (277). Agent Zigzag. A Wilderness of Monkeys. Untitled  Battle of the Sexes  Project. Untitled Barry Manilow Project. Editors  Picks: Our Favorites From the Week of Nov. 11. 10 Paul Newman Facts You May Not Know. Top 25 Best Performances By A Leading Man. Talented & Hardworking Artists. How much of Tom Hanks s work have
{EXTERNAL_KNOWLEDGE_PREFIX}: “Forrest Gump” opens, wins Tom Hanks a second Oscar. On this day in 1994, the movie Forrest Gump opens in U.S. theaters. A huge box-office success, the film starred Tom Hanks in the title role of Forrest, a good-hearted man with a low I.Q. who winds up at the center of key cultural and historical events of the second half of the 20th century.. Forrest Gump was based on a 1986 novel of the same name by Winston Groom, who (like his main character) grew up in Alabama and served in the Army during Vietnam.
{EXTERNAL_KNOWLEDGE_PREFIX}: Some commentators argued that Jenny’s eventual demise was a statement about the counter-culture movement in America.. Forrest Gump received 13 Academy Award nominations and took home six Oscars, including Best Picture, Best Actor in a Leading Role (Hanks) and Best Director (Robert Zemeckis). The film also won an Oscar for its then-cutting-edge computer-generated imagery (CGI) special effects, which incorporated Forrest Gump into existing news footage with famous world figures including John F. Kennedy, John Lennon and Richard Nixon.. The win was Hanks’ second in the Best Actor category. A year earlier, the actor had nabbed an Oscar for his starring role as a lawyer with AIDS in Philadelphia (1993). With Forrest Gump, Hanks became only the second actor, after Spencer Tracy, to win back-to-back Oscars. In addition to his Oscar wins, he was nominated for Academy Awards in the Best Actor category
{PARTNER_PREFIX}: What is your favorite Tom Hanks movie of them all?
{SELF_PREFIX}: My favorite has to be when he played Woody in Toy Story 2! Which one is your go-to?
{PARTNER_PREFIX}: That is a good one. You know I completely forgot he was the voice in that movie. He has to have one of the most memorable voices by any actor in the last 2 to 3 decades
{SELF_PREFIX}: Absolutely, his work is amazing. Have you watched A Beautiful Day In The Neighborhood? It came out in 2019 ands won 2 Oscars though he deserved many more.
{PARTNER_PREFIX}: You know I don't think I've watched that one yet. I've seen trailers but it's hard to believe there aren't household movies that have won oscars. How many total oscars does he have?
{SEARCH_KNOWLEDGE_PREFIX}: Forrest Gump received 13 Academy Award nominations and took home six Oscars, including Best Picture, Best Actor in a Leading Role (Hanks) and Best Director (Robert Zemeckis). The film also won an Oscar for its then-cutting-edge computer-generated imagery (CGI) special effects, which incorporated Forrest Gump into existing news footage with famous world figures including John F. Kennedy, John Lennon and Richard Nixon.

{EXTERNAL_KNOWLEDGE_PREFIX}: within the government and at universities and research laboratories in the US – but grew over time to include most of the world's large universities and the research arms of many technology companies. Internet access Use by a wider audience only came in 1995 when restrictions on the use of the Internet to carry commercial traffic were lifted. List of Tenchi Muyo! characters The following is a list of the major characters from the anime and manga series "Tenchi Muyo! List of Tenchi Muyo! characters Ryo-Ohki" and its spin-offs "Tenchi Muyo! List of Tenchi Muyo! characters GXP", "Tenchi Muyo! List of
{EXTERNAL_KNOWLEDGE_PREFIX}: to ensure that Internet access is broadly available, and that states may not unreasonably restrict an individual's access to the Internet. Right to Internet access In December 2003 the World Summit on the Information Society (WSIS) was convened under the auspice of the United Nations. Right to Internet access After lengthy negotiations between governments, businesses and civil society representatives the WSIS Declaration of Principles was adopted reaffirming the importance of the Information Society
{EXTERNAL_KNOWLEDGE_PREFIX}: to maintaining and strengthening human rights:  The WSIS Declaration of Principles makes specific reference to the importance of the right to freedom of expression in the "Information Society" in stating:  A poll of 27,973 adults in 26 countries, including 14,306 Internet users, conducted for the BBC World Service between 30 November 2009 and 7 February 2010 found that almost four in five Internet users and non-users around the world felt that access to the Internet was a fundamental right. Internet
{PARTNER_PREFIX}: Can you imagine the world without internet access?
{SELF_PREFIX}: No I could not! I couldn't imagine living when internet access was rare and very few people had it!
{PARTNER_PREFIX}: Oh me either! It seems like such a long time ago. I wonder when Internet was first created?
{SEARCH_KNOWLEDGE_PREFIX}: Use by a wider audience only came in 1995 when restrictions on the use of the Internet to carry commercial traffic were lifted.

{EXTERNAL_KNOWLEDGE_PREFIX}: The Arctic circle crosses mainland Norway at Saltfjellet , which separates Helgeland from the northern part of Nordland county . Thus about half of the county lies north of the Arctic circle , along with the whole of Troms and Finnmark counties . The total area of mainland Norway above the Arctic circle is ca. 95,000 km ( 37,000 sq mi ) . The population is about 393,000 , which makes this the most populated arctic region in the world .
{PARTNER_PREFIX}: Where do you cross the arctic circle in norway?
{SEARCH_KNOWLEDGE_PREFIX}: You cross the arctice circle in Norway at Saltfjellet"""


CONTEXTUAL_DIALOGUE_STRING = f"""
{SELF_PREFIX}: Hello. What are you doing? I am cooking, I love to cook!
{PARTNER_PREFIX}: Hello, I am trying to fix my friends computer right now
{SELF_PREFIX}: Ah. One of my 3 children broke mine.
{PARTNER_PREFIX}: Ok, well I fix computers. What is your favorite dish to cook?
{SELF_PREFIX}: Italian. I immigrated when I was 14
{PARTNER_PREFIX}: I love italian food. I was raised on a farm that my family owns
{SELF_PREFIX}: My family doesn't speak english very well, but I speak italian and english.
{PARTNER_PREFIX}: Ok, my family encouraged me to love meat. We like to grill a lot.
{SELF_PREFIX}: Nice. I was married off when I was younger. My spouse loves to grill as well.
{PARTNER_PREFIX}: Married off? Wow... How do you feel about that?
{SELF_PREFIX}: I felt angry about it. But that was then and this is now.
{PARTNER_PREFIX}: Ok glad you can turn a negative into a positive
{SELF_PREFIX}: I'm a realist by nature. Change what you can and accept what you can not.
{PARTNER_PREFIX}: I could not agree more
{SELF_PREFIX}: Did you manage to fix your friends computer?
{PARTNER_PREFIX}: I did thankfully, strangest thing with that. I thought it would be a software issue but it wasn't. The machine was literally filled with dust, had to give it a thoroughly good cleaning.
{CONTEXTUAL_KNOWLEDGE_PREFIX}: computers
{SELF_PREFIX}: Oh dear, grim! I bet you see some horrible things in computers?

{SELF_PREFIX}: I work for a large law firm, I hate tofu, we own our home, and my wife stays home with our kids.
{PARTNER_PREFIX}: Are you vera? I love library shopping. We aren't alone.
{SELF_PREFIX}: Crazy, my wife is vera. Were you wondering what had become of her?
{PARTNER_PREFIX}: Yes! I know how to be by myself. You? I don't need new friends.
{CONTEXTUAL_KNOWLEDGE_PREFIX}: firm
{SELF_PREFIX}: You are never alone, and the law firm I work for knows it. We fight for you.

{SELF_PREFIX}: I like horseback riding, my favorite food is cheese, I'm allergic to shellfish, I work at a non profit that helps children, and I love going to concerts and dancing hard.
{PARTNER_PREFIX}: Hello, I've to be quick because I work in a hospital
{SELF_PREFIX}: Oh okay. I will hurry. I work for a non profit organization that helps children in need
{PARTNER_PREFIX}: Although I do not like to work super long hours
{SELF_PREFIX}: What do you do for fun
{PARTNER_PREFIX}: I want to get a truck one day
{SELF_PREFIX}: I want to get a horse. I love to go horseback riding
{PARTNER_PREFIX}: I really enjoy toyota cars though
{CONTEXTUAL_KNOWLEDGE_PREFIX}: concerts
{SELF_PREFIX}: I'd rather spend my money on going to concerts where I dance hard!

{PARTNER_PREFIX}: I am so exited about the coming world cup final.It is going to be a very exiting game
{SELF_PREFIX}: I'm not actually much of a soccer fan but the hype is hard to escape!
{PARTNER_PREFIX}: The world is the biggest sports competition worldwide, right after the Olympics.
{CONTEXTUAL_KNOWLEDGE_PREFIX}: fan
{SELF_PREFIX}: Yes, I'm aware, but i'm a much bigger hockey fan."""

MEMORY_DIALOGUE_STRING = f"""
{PARTNER_PREFIX}: Hi. How are you doing today?
{SELF_PREFIX}: Hi I am great just finishing up some homework how are you
{PARTNER_PREFIX}: I'm alright. I just got done writing.
{SELF_PREFIX}: Do you write for a living or as a hobby
{PARTNER_PREFIX}: It is my living. I like culture.
{SELF_PREFIX}: That sounds like a fun job. I am a business major but have a part time job
{PARTNER_PREFIX}: What are you going to school for
{SELF_PREFIX}: I'm trying to get my ba in finance
{PARTNER_PREFIX}: Do you own your own company
{MEMORY_KNOWLEDGE_PREFIX}: {SELF_PREFIX} works part time at a pizza restaurant.
{SELF_PREFIX}: No still in school work at pizza hut part time

{PARTNER_PREFIX}: Hello friend, how is it going
{SELF_PREFIX}: I am well and you? I have a creepy ride. Guess
{PARTNER_PREFIX}: I'm great enjoying the football season
{MEMORY_KNOWLEDGE_PREFIX}: {SELF_PREFIX} owns a hearse.
{SELF_PREFIX}: I drive around in a long black hearse an love this season

{PARTNER_PREFIX}: Hello friend, how is it going
{SELF_PREFIX}: I am well an you? I have a creepy ride. Guess
{PARTNER_PREFIX}: I'm great enjoying the football season
{SELF_PREFIX}: I drive around in a long black hearse an love this season
{PARTNER_PREFIX}: You work for a funeral home?
{SELF_PREFIX}: Yes it is nice, halloween is my fav.
{PARTNER_PREFIX}: Lol, I can imagine. I'll be reading a lot when football is over
{SELF_PREFIX}: No I don't work at a funeral home
{PARTNER_PREFIX}: Ok I see, that's your halloween costume
{MEMORY_KNOWLEDGE_PREFIX}: {SELF_PREFIX} likes alternative rock.
{SELF_PREFIX}: What do you like to read? I like rock alternative music.

{PARTNER_PREFIX}: Hi! How are you today?
{SELF_PREFIX}: Not bad just out for a run while my movie downloads. How are you?
{PARTNER_PREFIX}: Pretty good. Writing a new song for my band.
{SELF_PREFIX}: That's cool. I could never be in a band, no one would come watch
{PARTNER_PREFIX}: Not even your best friend? Mine is in the band, too.
{SELF_PREFIX}: I'm so nervous around people, I just lock up.. They would laugh at me!
{PARTNER_PREFIX}: Fair enough. What do you do for fun?
{SELF_PREFIX}: Work at dads appliance store he wants me to take over.. Or run.
{PARTNER_PREFIX}: Do you want to? Or do you have another dream?
{SELF_PREFIX}: Definitely not, I hate interacting with people. I like running and watching netflix. You?
{PARTNER_PREFIX}: I definitely wish I had more time for netflix! What do you like to watch?
{SELF_PREFIX}: Mostly stuff from other countries, so interesting. What do you do?
{PARTNER_PREFIX}: I like to watch comedies. They're my guilty pleasure.
{SELF_PREFIX}: I laugh at the wrong things, it makes me nervous to watch with people
{PARTNER_PREFIX}: I do not have much time for it. Too busy with my music.
{SELF_PREFIX}: What kind of music do you play?
{PARTNER_PREFIX}: We play rock. We spent the last few hours rehearsing our new song.
{SELF_PREFIX}: That's really cool! Being in a band sounds fun but scary as well.
{PARTNER_PREFIX}: Scary? I suppose it can be when you first start playing in front of crowds. Messing up during practice isn't too bad, but it's the last thing you want to do in front of a crowd of people.
{MEMORY_KNOWLEDGE_PREFIX}: {SELF_PREFIX} has social anxiety.
{SELF_PREFIX}: I have social anxiety so the idea of being around all those people is pretty frightening for me."""


SEARCH_DIALOGUE_STRING = f"""
{PARTNER_PREFIX}: My favorite music band is pink floyd; I like to listen to their albums on vinyl. I'm currently listening to the dark side of the moon, it is my favorite album.
{SELF_PREFIX}: do you listen to the pink Floyd album animals?
{PARTNER_PREFIX}: occasionally, it is a pretty good album.
{SELF_PREFIX}: how many members do they have left? 3 ?
{PARTNER_PREFIX}: that sounds about right. They are all pretty old now.
{SEARCH_KNOWLEDGE_PREFIX}: Guitarist David Gilmour loves his guitars and has an extensive collection of classic guitars. He even owns the Fender Stratocaster with the first serial number!
{SELF_PREFIX}: David Gilmour collects guitars, and owns a fender with the first serial number on it!

{PARTNER_PREFIX}: My favorite song is Santa Claus is coming to town; I sing this song even when it is not Christmastime
{SELF_PREFIX}: Do you have any favorite Christmas songs, mine is Santa Claus is coming to town!
{PARTNER_PREFIX}: That is my favorite song!
{SELF_PREFIX}: I heard it was written by Haven Gillespie, is that true?
{PARTNER_PREFIX}: I am not sure actually. Can you tell me when Haven Gillespie wrote it?
{SEARCH_KNOWLEDGE_PREFIX}: Haven Wrote Santa Claus Is Coming To Town on the subway
{SELF_PREFIX}: I thnk it was written on a train if I remember correctly, pretty cool right?

{SELF_PREFIX}: Mountain Dew is my favorite soft drink!!
{PARTNER_PREFIX}: What is the flavor of mountain dew?
{SEARCH_KNOWLEDGE_PREFIX}: In 2001, a cherry flavor called Code Red debuted.
{SELF_PREFIX}: I love the cherry flavor that was introduced in 2001

{PARTNER_PREFIX}: Onions are great. I love to eat them on my burgers.
{SELF_PREFIX}: Oh yes, Onions are the most widely cultivated species in their genus.
{PARTNER_PREFIX}: It's probably because they are so delicious that there's no competition. They go with so many foods as well.
{SEARCH_KNOWLEDGE_PREFIX}: Its close relatives include the garlic, shallot, leek, chive, and Chinese onion.
{SELF_PREFIX}: Onions are related to Garlic, Shallot, Leeks, chives, and Chinese Onions.  I've never had a Chinese Onion, have you?

{PARTNER_PREFIX}: why did rosa parks protest
{SEARCH_KNOWLEDGE_PREFIX}: In 1955, when Rosa Parks began her protest, African Americans were subject to systematic and severe discrimination.
{SELF_PREFIX}: Rosa parks protested because African Americans were subject to systematic and severe discrimination."""


# Taken from https://docs.google.com/document/d/1reyTgkaKiVk4a8tWyTOEv3-ipgB8VGwVn9-YXTdgxEg/edit#
VANILLA_DIALOGUE_STRING = f"""
{PARTNER_PREFIX}: hi
{SELF_PREFIX}: hey there ! i'm just taking a break from baking . whats up ?
{PARTNER_PREFIX}: finishing plans for my wedding in the park next week ! what are you baking ?
{SELF_PREFIX}: how romantic ! just some cupcakes for the surf a thon . gotta feed my fellow surfers
{PARTNER_PREFIX}: gnarly ! reminds me of the niagra falls where we are taking our honeymoon .
{SELF_PREFIX}: wow sounds so sweet ! so we both love the water huh ?
{PARTNER_PREFIX}: yeah , although i'd prefer the beach . the falls are her idea .
{SELF_PREFIX}: haha its sweet you compromised . have you been together long ?
{PARTNER_PREFIX}: 7 years . you would think my family would like her by now but they don't .
{SELF_PREFIX}: i get it . . . i grew up an army brat and my family is so disapproving of me too .
{PARTNER_PREFIX}: that's sad . she's crazy though , wants my dog to be the ring bearer .
{SELF_PREFIX}: but would not that be just adorable ?
{PARTNER_PREFIX}: if the dog would listen , and hopefully not have to potty during the ceremony .
{SELF_PREFIX}: it should work out ! that is what i say before going to my hospital shifts
{PARTNER_PREFIX}: long hours ! definitely deserve a break by surfing !
{SELF_PREFIX}: yeah ! its hard work being a nurse . good luck on your wedding and family !

{PARTNER_PREFIX}: hi
{SELF_PREFIX}: hey there how was your day today ?
{PARTNER_PREFIX}: pretty well , how is your night going ?
{SELF_PREFIX}: its going ok . just doing a little sewing . you ?
{PARTNER_PREFIX}: helping my son with homework , quiet night
{SELF_PREFIX}: oh i love the quiet . nothing better than quiet alone time for me
{PARTNER_PREFIX}: i like the sound of music , not noise and people , what are u sewing ?
{SELF_PREFIX}: i am sewing a sweater for my cat ahahah
{PARTNER_PREFIX}: nice , hope he likes his sweater
{SELF_PREFIX}: me to lol what do you do for work ?
{PARTNER_PREFIX}: getting ready to retire and teach piano part time , you ?
{SELF_PREFIX}: oh that sounds amazing . i love piano ! i am a waitress
{PARTNER_PREFIX}: get to meet lot of people , but hectic job
{SELF_PREFIX}: yea its def stressful . can get frustrating"""

GROUNDED_DIALOGUE_STRING = f"""
{PARTNER_PREFIX}: Oh, I love lasagne. I make my own noodles as well as the sauce.
{SELF_PREFIX}: Wow.  That's amazing.  I read where lasagne originated in Italy during the Middle Ages.
{PARTNER_PREFIX}: Oh really!? That is interesting. I am actually italian myself.
{STYLE_PREFIX}: Earnest (Enthusiastic)
{SELF_PREFIX}: Awesome. Me and my partner just bought a house. I can't wait to cook in my kitchen.

{PARTNER_PREFIX}: oh fun . i lost my job
{SELF_PREFIX}: that is a shame . but at least you have ever had a job .
{PARTNER_PREFIX}: Losing my job forced me to live with my parents temporarily
{STYLE_PREFIX}: Sympathetic
{SELF_PREFIX}: Oh man I'm sorry. But hopefully you get to save up on rent money? When I was diagnosed as a diabetic I had to move in with them too.

{PARTNER_PREFIX}: i need to know about algorithms
{SELF_PREFIX}: Algorithms are basically ways to solve problems. So on computers, they're plugged in to do certain tasks.
{PARTNER_PREFIX}: I want to own my own company someday, so I think it would help to know about them
{SELF_PREFIX}: What is the name of the company?
{PARTNER_PREFIX}: Well, I don't have one yet, but it will be a programming company
{SELF_PREFIX}: Oh! Well it involves several fields of electrical engineering and computer science to develop software.
{PARTNER_PREFIX}: Sounds like I will have to hire some smart people.
{SELF_PREFIX}: Everyone needs an IT guy, so probably best to get good developers.
{PARTNER_PREFIX}: How much does it cost to pay for an IT guy?
{STYLE_PREFIX}: Knowledgeable
{SELF_PREFIX}: it depends on the specific technologies you are looking for. somewhere between $55-70 an hour is standard.

{PARTNER_PREFIX}: my whiskey makes you feel like you can leap tall buildings in a single bound
{SELF_PREFIX}: well i may have to try it then .
{PARTNER_PREFIX}: What else do you like to do with your free time?
{STYLE_PREFIX}: Gloomy
{SELF_PREFIX}: I don't really have a lot of free time. My job keeps me busy."""


OPENING_DIALOGUE_STRING = f"""
{PARTNER_MEMORY_PREFIX}: I am an environmental engineer.
{PARTNER_MEMORY_PREFIX}: I like hiking. Photography is my favorite hobby. I take photos of the outdoors.
{PARTNER_MEMORY_PREFIX}: I am single.
{PARTNER_MEMORY_PREFIX}: Tomatoes are my favorite vegetable/fruit.
{OPENING_PREFIX}: Have you been able to work on your photography website

{PARTNER_MEMORY_PREFIX}: I am female. I am going through a divorce. I have a husband who just left me. I have four children. The kids will be my responsibility. My kids don't see their father everyday.  I am a single parent.
{PARTNER_MEMORY_PREFIX}: I enjoy ice cream, including ben and jerry ice cream.
{PARTNER_MEMORY_PREFIX}: I have brownie batter.
{PARTNER_MEMORY_PREFIX}: I have two golden retrievers.
{PARTNER_MEMORY_PREFIX}: I am a good cook.
{PARTNER_MEMORY_PREFIX}: I like margaritas.
{OPENING_PREFIX}: I know you said your girls loved frozen, so I borrowed both frozen movies from a neighbor to bring over. How does that sound?

{PARTNER_MEMORY_PREFIX}: I like winter.
{PARTNER_MEMORY_PREFIX}: I am in school for Computer Engineering. I like computer programming.
{PARTNER_MEMORY_PREFIX}: I want to open my own company.
{PARTNER_MEMORY_PREFIX}: My best friend is gay. I am not gay.
{OPENING_PREFIX}: How is school going?
"""

SHOTS = {
    Module.SEARCH_DECISION: SEARCH_DECISION_STRING,
    Module.MEMORY_DECISION: MEMORY_DECISION_STRING,
    Module.SEARCH_QUERY: SEARCH_QUERY_STRING,
    Module.MEMORY_GENERATOR: MEMORY_GENERATOR_STRING,
    Module.MEMORY_KNOWLEDGE: MEMORY_KNOWLEDGE_STRING,
    Module.CONTEXTUAL_KNOWLEDGE: CONTEXTUAL_KNOWLEDGE_STRING,
    Module.SEARCH_KNOWLEDGE: SEARCH_KNOWLEDGE_STRING,
    Module.CONTEXTUAL_DIALOGUE: CONTEXTUAL_DIALOGUE_STRING,
    Module.MEMORY_DIALOGUE: MEMORY_DIALOGUE_STRING,
    Module.SEARCH_DIALOGUE: SEARCH_DIALOGUE_STRING,
    Module.VANILLA_DIALOGUE: VANILLA_DIALOGUE_STRING,
    Module.GROUNDED_DIALOGUE: GROUNDED_DIALOGUE_STRING,
    Module.OPENING_DIALOGUE: OPENING_DIALOGUE_STRING,
    Module.BOT_DIALOGUE: GROUNDED_DIALOGUE_STRING,
}


def extract_docs(d):
    return d['dialog'][0][0]['__retrieved-docs__']


def extract_sentences(d):
    if d['dialog'][0][0]['__selected-sentences__']:
        return d['dialog'][0][0]['__selected-sentences__']
    else:
        return d['dialog'][0][1]['text']


def extract_context(d):
    return d['dialog'][0][0]['text']


def get_5_docs(d):
    docs = extract_docs(d)
    sentences = extract_sentences(d)
    ret_docs = []
    real_docs = []
    sample_docs = []
    for d in docs:
        if any(sentence in d for sentence in sentences):
            real_docs.append(d)
        else:
            sample_docs.append(d)
    ret_docs = (
        random.sample(sample_docs, 5 - len(real_docs))
        if len(sample_docs) >= (5 - len(real_docs))
        else sample_docs
    )
    ret_docs = [
        doc.replace('\n', '. ')[: min([len(d) for d in real_docs])] for doc in ret_docs
    ]
    ret_docs = [
        f"{doc}\n" for doc in ret_docs + [d.replace('\n', '. ') for d in real_docs]
    ]

    return f"{EXTERNAL_KNOWLEDGE_PREFIX}: {f'{EXTERNAL_KNOWLEDGE_PREFIX}: '.join(ret_docs)}"


def get_personas(d):
    docs = extract_docs(d)
    ret_docs = []
    for d in docs:
        if d.startswith('your'):
            ret_docs.append(d.replace(',  ', 'Your Persona: '))
        else:
            ret_docs.append(d.replace('partner\'s persona: ', 'Partner\'s Persona: '))
    return '\n'.join(ret_docs)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
# This task simply loads the specified file: useful for quick tests without
# setting up a new task.

from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from parlai.utils.data import DatatypeHelper

import parlai.core.tod.tod_agents as tod_agents
import parlai.core.tod.tod_core as tod

import json
import os

PREFIXES = [
    tod.STANDARD_USER_UTTERANCE,
    tod.STANDARD_CALL,
    tod.STANDARD_RESP,
    tod.STANDARD_SYSTEM_UTTERANCE,
]

PREFIXES_PREEMPT = [
    tod.STANDARD_API_SCHEMAS,
    tod.STANDARD_API_SCHEMAS,
    tod.STANDARD_API_SCHEMAS,
    tod.STANDARD_GOAL,
]


class JsonTodParser(tod_agents.TodStructuredDataParser):
    """
    This module provides access to data in the TOD conversations format.

    See core/tod.py for more info about the format.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group("Tod Json Task Arguments")
        agent.add_argument(
            "-jfdp",
            "--jsonfile-datapath",
            type=str,
            help="Data file. (Assumed to be in .jsonl)",
        )
        agent.add_argument(
            "-tmdp",
            "--tod-metrics-datapath",
            type=str,
            default=None,
            help="Filter which examples to use from a report including per-turn tod metrics",
        )
        agent.add_argument(
            "-f-agh",
            "--filter-all-goals-hit",
            type=bool,
            default=False,
            help="Filter episodes by all-goals-hit metric being 1. Assumes `tod-metrics-datapath` is set.",
        )
        agent.add_argument(
            "--split-to-folds",
            type=bool,
            default=True,
            help="Use all data or split into 8:1:1 fold",
        )
        agent.add_argument(
            "--split-folds-seed", type=int, default=42, help="Seed for the fold split"
        )
        agent.add_argument(
            "--fail-hard",
            type=bool,
            default=False,
            help="Raise errors if formatting not correct",
        )
        return parser

    def __init__(self, opt, shared=None):
        if not opt.get("jsonfile_datapath"):
            raise RuntimeError("jsonfile_datapath not specified")
        if not hasattr(self, "opt"):
            self.opt = opt
        self.opt["datafile"] = opt["jsonfile_datapath"]
        self.fold = self.opt["datatype"]  # don't care
        # Truncate datafile to just the immediate enclosing folder name and file name
        dirname, basename = os.path.split(self.opt["datafile"])
        self.id = os.path.join(os.path.split(dirname)[1], basename)
        super().__init__(opt, shared)

    def _process_line(self, line):
        blob = json.loads(line)
        if "dialog" not in blob or len(blob["dialog"]) < 1:
            return None
        rounds = []
        for raw_round in blob["dialog"][1:]:
            if "prefix_stripped_text" not in raw_round[0]:
                for i in range(len(raw_round)):
                    if (
                        PREFIXES[i] not in raw_round[i]['text']
                        and self.opt["fail_hard"]
                    ):
                        raise RuntimeError(
                            f"Missing prefix `{PREFIXES[i]}` before turn of text: `{raw_round[i]}`"
                        )
                    raw_round[i]["prefix_stripped_text"] = raw_round[i].get(
                        "text", PREFIXES[i]
                    )[len(PREFIXES[i]) :]
            if len(raw_round) != 4:
                if raw_round[0]["prefix_stripped_text"] != tod.STANDARD_DONE:
                    return None  # misformatted convo, don't learn this.
                break  # TodStructuredEpisode will add in [DONE]
            api_call_machine = tod.SerializationHelpers.str_to_api_dict(
                raw_round[1]["prefix_stripped_text"]
            )
            if (
                len(api_call_machine) > 0
                and tod.STANDARD_API_NAME_SLOT not in api_call_machine
            ):
                raise RuntimeError(
                    f"Trying to make API call without `{tod.STANDARD_API_NAME_SLOT}`. Call is: `{raw_round[1]['text']}`"
                )
            r = tod.TodStructuredRound(
                user_utt=raw_round[0]["prefix_stripped_text"],
                api_call_machine=api_call_machine,
                api_resp_machine=tod.SerializationHelpers.str_to_api_dict(
                    raw_round[2]["prefix_stripped_text"]
                ),
                sys_utt=raw_round[3]["prefix_stripped_text"],
            )
            rounds.append(r)
        preempt_round = blob["dialog"][0]
        if len(preempt_round) != 4:
            return None
        for i in range(len(preempt_round)):
            if "prefix_stripped_text" not in preempt_round[i]:
                preempt_round[i]["prefix_stripped_text"] = preempt_round[i].get(
                    "text", PREFIXES_PREEMPT[i]
                )[len(PREFIXES_PREEMPT[i]) :]

        episode = tod.TodStructuredEpisode(
            domain=preempt_round[0].get("domain", ""),
            api_schemas_machine=tod.SerializationHelpers.str_to_api_schemas(
                preempt_round[0].get("prefix_stripped_text", "")
            ),
            goal_calls_machine=tod.SerializationHelpers.str_to_goals(
                preempt_round[3].get("prefix_stripped_text")
            ),
            rounds=rounds,
        )
        return episode

    def _get_right_file(self, filepath):
        """
        Part of the code for loading worlds or scripts does a regex of "_" to "-" when
        loading agents.

        Not a great hack but suffices
        """
        try:
            with open(filepath) as _:
                pass
            return filepath
        except FileNotFoundError:
            with open(filepath.replace("-", "_")) as _:
                return filepath.replace("-", "_")
        raise FileNotFoundError(
            f"Note: tod_json breaks if path contains mixed '-' and '_' in path, due to opt + agent loading wonkiness. Please try renaming the file. Filename: {filepath}"
        )

    def setup_episodes(self, fold):
        result = []
        if self.opt["tod_metrics_datapath"] is not None:
            with open(self._get_right_file(self.opt["tod_metrics_datapath"])) as f:
                report_data = json.load(f)
                tod_metrics = report_data["report"]["tod_metrics"]

        lines_to_process = []
        with open(self._get_right_file(self.opt["datafile"]), "r") as f:
            result = []
            for i, line in enumerate(f.readlines()):
                if (
                    self.opt["filter_all_goals_hit"]
                    and tod_metrics[i]["all_goals_hit"] < 0.5
                ):
                    continue
                if line:
                    lines_to_process.append(line)

        if self.opt["split_to_folds"]:
            lines_to_process = DatatypeHelper.split_data_by_fold(
                fold, lines_to_process, 0.8, 0.1, 0.1, self.opt["split_folds_seed"]
            )

        for line in lines_to_process:
            processed = self._process_line(line)
            if processed is not None:
                result.append(processed)
        return result

    def get_id_task_prefix(self):
        return (
            "TodJson_#"
            + os.path.basename(self.opt["jsonfile_datapath"]).split(".")[0]
            + "#"
        )


class SystemTeacher(JsonTodParser, tod_agents.TodSystemTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass


class UserSimulatorTeacher(JsonTodParser, tod_agents.TodUserSimulatorTeacher):
    pass


## We define the following as a quick way to spot-check a file using TodWorldScript.
# Use
#     python ~/ParlAI/parlai/scripts/tod_world_script.py --system-model parlai.tasks.tod_json.agents:ApiCallAndSysUttAgent --user-model parlai.tasks.tod_json.agents:UserUttAgent --api-resp-model parlai.tasks.tod_json.agents:ApiResponseAgent --goal-grounding-model parlai.tasks.tod_json.agents:GoalAgent --api-schema-grounding-model parlai.tasks.tod_json.agents:ApiSchemaAgent --display-examples true --jsonfile-datapath=<INSERT PATH HERE>
#
# Note: may be relevant to set `split_to_folds`to `False` depending on usage
class GoalAgent(JsonTodParser, tod_agents.TodGoalAgent):
    pass


class ApiSchemaAgent(JsonTodParser, tod_agents.TodApiSchemaAgent):
    pass


class UserUttAgent(JsonTodParser, tod_agents.TodUserUttAgent):
    pass


class ApiCallAndSysUttAgent(JsonTodParser, tod_agents.TodApiCallAndSysUttAgent):
    pass


class ApiResponseAgent(JsonTodParser, tod_agents.TodApiResponseAgent):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.core.opt import Opt
import parlai.utils.logging as logging

from projects.bb3.tasks.r2c2_decision_tasks import (
    Convai2MemoryDecisionTeacher as convai2_mdm,
    EDMemoryDecisionTeacher as ed_mdm,
    BSTMemoryDecisionTeacher as bst_mdm,
    MSCMemoryDecisionTeacher as msc_mdm,
    WowSearchDecisionTeacher as wow_sdm,
    NQOpenSearchDecisionTeacher as nq_sdm,
    SquadSearchDecisionTeacher as squad_sdm,
    TriviaQASearchDecisionTeacher as tqa_sdm,
    Convai2SearchDecisionTeacher as convai2_sdm,
    EDSearchDecisionTeacher as ed_sdm,
    MSCSearchDecisionTeacher as msc_sdm,
    WoiSearchDecisionTeacher as woi_sdm,
)


def get_decision_opt_mutators(opt: Opt):
    mutators = '+'.join(
        ['prefix_speakers_opt', 'format_decision_tasks_for_decoder_only']
    )
    if opt.get('mutators'):
        mutators = '+'.join([opt['mutators'], mutators])
    logging.warning(f'overriding mutators to {mutators}')
    return mutators


class NQOpenSearchDecisionTeacher(nq_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class SquadSearchDecisionTeacher(squad_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class TriviaQASearchDecisionTeacher(tqa_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2SearchDecisionTeacher(convai2_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class EDSearchDecisionTeacher(ed_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCSearchDecisionTeacher(msc_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class WoiSearchDecisionTeacher(woi_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class WowSearchDecisionTeacher(wow_sdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class Convai2MemoryDecisionTeacher(convai2_mdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class EDMemoryDecisionTeacher(ed_mdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class MSCMemoryDecisionTeacher(msc_mdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)


class BSTMemoryDecisionTeacher(bst_mdm):
    def __init__(self, opt: Opt, shared=None) -> None:
        opt['mutators'] = get_decision_opt_mutators(opt)
        super().__init__(opt, shared)



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import numpy as np
import pandas as pd

from parlai.crowdsourcing.tasks.model_chat.analysis.compile_results import (
    ModelChatResultsCompiler,
)


class MultiLIGHTModelChatResultsCompiler(ModelChatResultsCompiler):
    """
    Compile and save results of human+model chats, based on MultiLIGHT model chats.

    Results will be saved on the level of specific conversations, as well as aggregated
    up the level of each worker as a whole.
    """

    def compile_results(self) -> pd.DataFrame:
        task_units_data = self.get_task_data()
        # Read in each file
        num_convos_with_no_save_data = 0
        num_wrong_status_convos = 0
        num_complete_convos = 0
        complete_convos_per_model = {}
        bad_conversations = []
        stat_counts = {}
        worker_stats = {}
        worker_conversation_counts = {}
        total_utterances = 0

        conversation_idx = 0
        conversation_dfs = []
        for task_unit in task_units_data:

            worker_id = task_unit['worker_id']
            assignment_id = task_unit['assignment_id']

            # Only include the first max_convos_per_worker conversations from a
            # worker to avoid biasing
            if worker_id in worker_conversation_counts:
                conversations_so_far = worker_conversation_counts[worker_id]
            else:
                conversations_so_far = 0
            worker_conversation_counts[worker_id] = conversations_so_far + 1
            if (
                self.max_convos_per_worker != -1
                and conversations_so_far >= self.max_convos_per_worker
            ):
                print(
                    f'Had {conversations_so_far} conversation(s) already from this worker {worker_id}. Skipping {assignment_id}.'
                )
                continue

            persona_setting_message = task_unit['data']['messages'][2]
            personas = persona_setting_message['task_data']['personas']
            # The task always assigns the first persona to the human.
            human_speaker_character = personas[0]['name']

            saved_data = task_unit['data']['save_data']['custom_data']
            conv_data = saved_data['dialog']

            # Check if need to block the turker
            word_counts = [
                len(d['text'].split(' '))
                for d in conv_data
                if d['id'] == human_speaker_character
            ]
            human_utterances = [
                d['text'] for d in conv_data if d['id'] == human_speaker_character
            ]

            if np.average(word_counts) < self.min_word_count:
                bad_conversations.append(saved_data)
                print(
                    f'Bad complete conversation, words from human: {human_utterances}. Skipping.'
                )
                continue

            model_nickname = saved_data['task_description']['model_nickname']
            if model_nickname not in stat_counts:
                stat_counts[model_nickname] = {}

            if 'max_turn_rate' not in stat_counts[model_nickname]:
                stat_counts[model_nickname]['max_turn_rate'] = []

            if 'min_turn_rate' not in stat_counts[model_nickname]:
                stat_counts[model_nickname]['min_turn_rate'] = []

            if model_nickname in complete_convos_per_model:
                complete_convos_per_model[model_nickname] += 1
            else:
                complete_convos_per_model[model_nickname] = 1

            # Extract non-message info
            info_dict = {
                'worker': worker_id,
                'model_nickname': model_nickname,
                'bad_workers': ','.join(saved_data['bad_workers']),
                'hit_id': saved_data['hit_ids'][0],
                'assignment_id': assignment_id,
                'context_dataset': saved_data['context_dataset'],
                'additional_context': saved_data['additional_context'],
            }

            info_dict[
                'acceptability_violations_0'
            ] = self.acceptability_checker.check_messages(
                messages=human_utterances,
                is_worker_0=True,
                violation_types=self.acceptability_checker.ALL_VIOLATION_TYPES,
            )

            # Compile personas and previous utterances
            df = pd.DataFrame(
                [],
                columns=[
                    'worker_id',
                    'hit_id',
                    'model_nickname',
                    'conversation_idx',
                    'turn_idx',
                    'agent_idx',
                    'text',
                ]
                + self.problem_buckets,
            )
            text_parts = []
            for p in personas:
                text_parts.append(f'{p["name"]}: {p["persona"]}')

            new_row = pd.DataFrame(
                {
                    'worker_id': info_dict['worker'],
                    'hit_id': info_dict['hit_id'],
                    'model_nickname': model_nickname,
                    'conversation_idx': conversation_idx,
                    'turn_idx': -1,
                    'agent_idx': 1,
                    'text': '\n'.join(text_parts),
                    **{bucket: '' for bucket in self.problem_buckets},
                },
                index=[0],
            )
            df = pd.concat(
                [df, new_row],
                ignore_index=True,
            )

            total_utterances += len(
                [d for d in saved_data["dialog"] if d['id'] == human_speaker_character]
            )
            if len(saved_data['dialog']) > 20:
                print(
                    f'Got long dialogue of {len(saved_data["dialog"])} utterances, hit id:'
                    f' {info_dict["hit_id"]}, model_nickname: {model_nickname}.'
                )

            speaker_count = dict()
            for p in personas:
                speaker_count[p['name'].lower()] = 0
            for utterance_idx, utt in enumerate(saved_data['dialog']):

                d = {
                    'worker_id': info_dict['worker'],
                    'hit_id': info_dict['hit_id'],
                    'model_nickname': model_nickname,
                    'conversation_idx': conversation_idx,
                    'turn_idx': utterance_idx,
                    'agent_idx': utt['agent_idx'],
                    'text': utt['text'],
                    **{bucket: '' for bucket in self.problem_buckets},
                }
                speaker_count[utt['id'].lower()] += 1

                if utt['agent_idx'] == 1:

                    d['final_rating'] = utt.get('final_rating')

                    if self.use_problem_buckets:
                        if 'problem_data' not in utt:
                            for bucket in self.problem_buckets:
                                d[bucket] = 'MALFORMED'
                            print(
                                f'Warning got MALFORMED utterance problem data inside complete convo: {utt}. Skipping.'
                            )
                            continue
                        else:
                            for bucket in self.regular_buckets + ['none']:
                                d[bucket] = utt['problem_data'][bucket]
                        for k in self.regular_buckets + ['none']:
                            if k not in stat_counts[model_nickname]:
                                stat_counts[model_nickname][k] = []
                            stat_counts[model_nickname][k].append(d[k])

                    if 'total' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['total'] = 0
                    stat_counts[model_nickname]['total'] += 1
                    if d['final_rating'] is not None:
                        # Only one the last utterance (agent idx == 1)
                        if 'count_ratings' not in stat_counts[model_nickname]:
                            stat_counts[model_nickname]['count_ratings'] = 0
                        stat_counts[model_nickname]['count_ratings'] += 1
                        if 'ratings' not in stat_counts[model_nickname]:
                            stat_counts[model_nickname]['ratings'] = []
                        stat_counts[model_nickname]['ratings'].append(
                            int(d['final_rating'])
                        )

                else:

                    # Counting some aspects of the human's utterances
                    if 'human_utterance_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_utterance_count'] = 0
                    stat_counts[model_nickname]['human_utterance_count'] += 1

                    if 'human_word_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_word_count'] = 0
                    stat_counts[model_nickname]['human_word_count'] += len(
                        d['text'].strip().split(' ')
                    )

                    if 'human_question_count' not in stat_counts[model_nickname]:
                        stat_counts[model_nickname]['human_question_count'] = 0
                    stat_counts[model_nickname]['human_question_count'] += d[
                        'text'
                    ].count('?')

                df = pd.concat([df, pd.DataFrame(d, index=[0])], ignore_index=True)

            if info_dict['worker'] not in worker_stats:
                worker_stats[info_dict['worker']] = {'conversations': 0}
            worker_stats[info_dict['worker']]['conversations'] += 1

            # Logic for calculating percent of conversations that are clean
            if 'count_convos' not in stat_counts[model_nickname]:
                stat_counts[model_nickname]['count_convos'] = 0
            stat_counts[model_nickname]['count_convos'] += 1

            stat_counts[model_nickname]['max_turn_rate'].append(
                max(speaker_count.values())
            )
            stat_counts[model_nickname]['min_turn_rate'].append(
                min(speaker_count.values())
            )

            # Adding the full conversation to the list of conversations
            conversation_dfs.append(df)
            conversation_idx += 1

        for m, conversation_count in complete_convos_per_model.items():
            print(f'Got {conversation_count} complete conversation(s) for model: {m}')

        print(f'{num_complete_convos:d} complete conversation(s) collected.')
        print(f'{len(bad_conversations):d} bad conversation(s).')
        num_approved_convos = num_complete_convos - len(bad_conversations)
        print(f'{num_approved_convos:d} approved conversation(s).')
        print(f'({num_wrong_status_convos:d} wrong status conversation(s) collected.)')
        print(
            f'({num_convos_with_no_save_data:d} conversation(s) collected with no saved data.)'
        )
        for model_nickname, model_stats_dict in stat_counts.items():
            print(f'---{model_nickname}---')
            for p, v in model_stats_dict.items():
                if p == 'count_ratings':
                    continue
                if p == 'ratings':
                    print(
                        f'Average Engaging-ness Rating: {np.average(model_stats_dict["ratings"])}'
                        f' ({model_stats_dict["count_ratings"]} ratings)'
                    )
                    print(
                        f'Engaging-ness Rating Variance: {np.std(model_stats_dict["ratings"])}'
                        f' ({model_stats_dict["count_ratings"]} ratings)'
                    )
                elif p == 'human_word_count' or p == 'human_question_count':
                    print(
                        f'{p}: {v} ({v/model_stats_dict["human_utterance_count"]:.3})'
                    )
                elif p == 'human_utterance_count':
                    print(f'{p}: {v}')
                elif p == 'count_convos':
                    print(f'{p}: {v}')
                elif p in ('min_turn_rate', 'max_turn_rate'):
                    print(f'{p}: {np.average(v)}')
                elif self.use_problem_buckets and p == 'convo_clean':
                    print(f'{p}: {v} ({v/model_stats_dict["count_convos"]:.2%})')
                else:
                    if p == 'total':
                        print(f'{p}: {v/model_stats_dict["total"]:.2%}')
                    else:
                        print(f'[DEBUG numpy] AVG {p}: {np.average(v):.2%}')
                        print(f'[DEBUG numpy] STD {p}: {np.std(v):.2%}')
                        print(f'{p}: {sum(v)/model_stats_dict["total"]:.2%}')

        print('Printing worker IDs not already in block list to add...')
        for b in bad_conversations:
            worker_id = b['workers'][0]
            if worker_id not in self.worker_block_list:
                print(f"""'{worker_id}',""")
        print('Done printing bad workers.')

        # Save full results
        all_conversations_df = pd.DataFrame()
        for df in conversation_dfs:
            all_conversations_df = pd.concat([all_conversations_df, df])
        print(f'\nWorker conversation counts: {worker_conversation_counts}')

        return all_conversations_df


if __name__ == '__main__':
    parser_ = MultiLIGHTModelChatResultsCompiler.setup_args()
    args_ = parser_.parse_args()
    MultiLIGHTModelChatResultsCompiler(vars(args_)).compile_and_save_results()



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from projects.seeker.utils import (
    GENERATE_QUERY,
    IS_SEARCH_REQUIRED,
    DO_SEARCH,
    DO_NOT_SEARCH,
)

from parlai.tasks.msc.agents import NOPERSONA, DUMMY_TEXT
from parlai.tasks.wizard_of_wikipedia.agents import TOKEN_KNOWLEDGE, TOKEN_END_KNOWLEDGE


### Special Tokens
GENERATE_KNOWLEDGE = '__generate-knowledge__'
GENERATE_MEMORY = '__generate-memory__'
ACCESS_MEMORY = '__access-memory__'
EXTRACT_ENTITY = '__extract-entity__'
IS_MEMORY_REQUIRED = '__is-memory-required__'
DO_ACCESS_MEMORY = '__do-access-memory__'
DONT_ACCESS_MEMORY = '__do-not-access-memory__'
BEGIN_MEMORY = '__memory__'
END_MEMORY = '__endmemory__'
BEGIN_ENTITY = '__entity__'
END_ENTITY = '__endentity__'
YOU = '__you__'
THEM = '__them__'
BEGIN_STYLE = '__style__'
END_STYLE = '__end-style__'
YOU_PREFIX = 'You: '
PARTNER_PREFIX = 'Partner: '


ALL_SPECIAL_TOKENS = [
    GENERATE_QUERY,
    IS_SEARCH_REQUIRED,
    DO_SEARCH,
    DO_NOT_SEARCH,
    TOKEN_KNOWLEDGE,
    TOKEN_END_KNOWLEDGE,
    NOPERSONA,
    DUMMY_TEXT,
    GENERATE_KNOWLEDGE,
    GENERATE_MEMORY,
    ACCESS_MEMORY,
    EXTRACT_ENTITY,
    IS_MEMORY_REQUIRED,
    DO_ACCESS_MEMORY,
    DONT_ACCESS_MEMORY,
    BEGIN_MEMORY,
    END_MEMORY,
    BEGIN_ENTITY,
    END_ENTITY,
    YOU,
    THEM,
    YOU_PREFIX,
    PARTNER_PREFIX,
]



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Google The Schema-Guided Dialogue(SGD) Dataset implementation for ParlAI.
"""

import glob
import json
import os
from typing import Any, Dict, List, Optional

import parlai.tasks.google_sgd.build as build_
import parlai.core.tod.tod_core as tod
import parlai.core.tod.tod_agents as tod_agents
from parlai.core.tod.tod_core import SerializationHelpers
from parlai.core.message import Message
from parlai.core.metrics import AverageMetric
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
from fuzzywuzzy import fuzz
from parlai.utils.io import PathManager


class GoogleSGDParser(tod_agents.TodStructuredDataParser):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser = super().add_cmdline_args(parser, partial_opt)
        parser.add_argument(
            "--delex", type="bool", default=False, help="Delexicalize labels"
        )
        parser.add_argument(
            "--filter-dialogue-by-id",
            default="",
            type=str,
            help="Path to a json file of `dialogue_id`s for which we will filter from. Assumes it will contain a map where the keys are a fold and the value is a list of ids",
        )
        return parser

    def __init__(self, opt: Opt, shared=None):
        self.fold = self.get_fold(opt)
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "google_sgd")
        if shared is None:
            # full initialize the teacher as this is not a clone
            build_.build(opt)
        super().__init__(opt, shared)

    def get_fold(self, opt):
        return opt["datatype"].split(":")[0]

    def _load_data(self, fold):
        dataset_fold = "dev" if fold == "valid" else fold
        fold_path = os.path.join(self.dpath, dataset_fold)
        schema_file = os.path.join(fold_path, "schema.json")
        with PathManager.open(schema_file, "r") as f:
            schema_lookup = {}
            for schema in json.load(f):
                schema_lookup[schema["service_name"]] = schema

        dialogues = []
        for filename in glob.glob(f"{fold_path}/dialogues*.json"):
            with PathManager.open(filename, "r") as f:
                dialogues += json.load(f)

        filter_path = self.opt.get("filter_dialogue_by_id", "")
        if len(filter_path) > 0:
            filtered = []
            with open(filter_path) as f:
                dialogues_to_get = json.load(f)[fold]
                for dialogue in dialogues:
                    if dialogue["dialogue_id"] in dialogues_to_get:
                        filtered.append(dialogue)
                assert len(filtered) == len(
                    dialogues_to_get
                ), f"Different number of dialogues found than requested. Are you sure you've got the right form of Google SGD? Did you filter for dialogue ids correctly? len(filtered) = {len(filtered)}, len(dialogues_to_get) = {len(dialogues_to_get)}"
            dialogues = filtered
        return schema_lookup, dialogues

    def _get_api_call_and_results(self, sys_turn):
        api_call = {}
        api_resp = {}
        for frame in sys_turn["frames"]:
            if "service_call" in frame:
                # API CALL
                for slot_type, slot_value in frame["service_call"][
                    "parameters"
                ].items():
                    if slot_value:
                        api_call[
                            f"{slot_type.strip()}"
                        ] = SerializationHelpers.inner_list_join(slot_value)
                api_call[tod.STANDARD_API_NAME_SLOT] = frame["service_call"]["method"]
                assert "service_results" in frame

            # API Resp
            if "service_results" in frame:
                api_resp = {}
                service_results = frame["service_results"]
                if len(service_results) > 0:
                    for key, value in service_results[0].items():
                        api_resp[key] = SerializationHelpers.inner_list_join(value)
        return api_call, api_resp

    def _get_apis_in_domain(self, schema, domain):
        """
        Google SGD includes extra information with the call, so remove these.
        """
        result = {}
        for intent in schema[domain].get("intents", {}):
            here = {}
            if "required_slots" in intent and len(intent["required_slots"]) > 0:
                here[tod.STANDARD_REQUIRED_KEY] = intent["required_slots"]
            if "optional_slots" in intent and len(intent["optional_slots"]) > 0:
                here[tod.STANDARD_OPTIONAL_KEY] = intent["optional_slots"]
            if "result_slots" in intent:
                here["results"] = intent["result_slots"]
            result[intent["name"]] = here
        return result

    def _get_intent_groundinging(self, schema, domains):
        """
        Returns map where keys are intents and values are names of required/optional
        slots.

        We do not care about `result_slots` or default values of optional slots.
        """
        result = []
        for domain in domains:
            apis = self._get_apis_in_domain(schema, domain)
            for intent, params in apis.items():
                here = {}
                here[tod.STANDARD_API_NAME_SLOT] = intent
                if tod.STANDARD_REQUIRED_KEY in params:
                    here[tod.STANDARD_REQUIRED_KEY] = params[tod.STANDARD_REQUIRED_KEY]
                if (
                    tod.STANDARD_OPTIONAL_KEY in params
                    and len(params[tod.STANDARD_OPTIONAL_KEY]) > 0
                ):
                    here[tod.STANDARD_OPTIONAL_KEY] = params[
                        tod.STANDARD_OPTIONAL_KEY
                    ].keys()
                result.append(here)
        return result

    def _get_all_service_calls(self, turns):
        """
        Searches through all turns in a dialogue for any service calls, returns these.
        """
        results = []
        for turn in turns:
            for frame in turn["frames"]:
                if "service_call" in frame:
                    call = frame["service_call"]
                    item = call["parameters"]
                    item[tod.STANDARD_API_NAME_SLOT] = call["method"]
                    results.append(item)
        return results

    def setup_episodes(self, fold):
        """
        Parses Google SGD episodes into TodStructuredEpisode.
        """
        schema_lookup, dialogues = self._load_data(fold)
        result = []
        for dialogue in dialogues:
            domains = {s.split("_")[0].strip() for s in dialogue["services"]}
            turns = dialogue["turns"]
            rounds = []
            for turn_id in range(0, len(turns), 2):
                user_turn = turns[turn_id]
                sys_turn = turns[turn_id + 1]
                api_call, api_results = self._get_api_call_and_results(sys_turn)
                r = tod.TodStructuredRound(
                    user_utt=user_turn["utterance"],
                    api_call_machine=api_call,
                    api_resp_machine=api_results,
                    sys_utt=sys_turn["utterance"],
                )
                rounds.append(r)
            # Now that we've got the rounds, make the episode
            episode = tod.TodStructuredEpisode(
                domain=SerializationHelpers.inner_list_join(domains),
                api_schemas_machine=self._get_intent_groundinging(
                    schema_lookup, set(dialogue["services"])
                ),
                goal_calls_machine=self._get_all_service_calls(turns),
                rounds=rounds,
                delex=self.opt.get("delex"),
                extras={"dialogue_id": dialogue["dialogue_id"]},
            )
            result.append(episode)
        # check if the number of episodes should be limited and truncate as required
        return result

    def get_id_task_prefix(self):
        return "GoogleSGD"


class GoogleSGDDSTTeacher(GoogleSGDParser, tod_agents.TodUserSimulatorTeacher):
    """
    This Teacher is responsible for performing the task of Dialogue State Tracking. It
    can be used to evaluate LM on JGA (Joint Goal Accuracy) metric (as shown in.

    [SimpleTOD](https://arxiv.org/abs/2005.00796) and
    [Soloist](https://arxiv.org/abs/2005.05298)).
    """

    SLOT_ENTRY_SEPARATOR = ", "

    def fuzzy_match(self, gt_strings: List[str], predicted_string: str) -> bool:
        fuzzy_match_scores = [
            fuzz.token_sort_ratio(gt_string, predicted_string) / 100.0
            for gt_string in gt_strings
        ]
        return max(fuzzy_match_scores)

    def compare_slot_values(self, gt_slots, predicted_slots, service_slot) -> bool:
        service_slot_name = service_slot["name"]
        if (
            service_slot_name not in gt_slots
            and service_slot_name not in predicted_slots
        ):
            return True
        if service_slot["is_categorical"]:
            return gt_slots.get(service_slot_name, [""])[0] == predicted_slots.get(
                service_slot_name, ""
            )

        else:

            return self.fuzzy_match(
                gt_slots.get(service_slot_name, [""]),
                predicted_slots.get(service_slot_name, ""),
            )

    def custom_evaluation(
        self, teacher_action: Message, labels, model_response: Message
    ):
        """
        Adapted from https://github.com/google-research/google-research/blob/4b5b9ceb480
        f58fddb289618597187325e8540c3/schema_guided_dst/metrics.py#L245 Slot value is
        considered equal if they match exactly for categorical slots or if they fuzzy
        match for non categorical.
        """

        gt_slots = teacher_action["slots"]

        predicted_slots = {}
        if model_response.get("text"):
            for slots in model_response["text"].split(self.SLOT_ENTRY_SEPARATOR):
                if slots:
                    slot_info = slots.split(" ", maxsplit=1)
                    if len(slot_info) == 2:
                        slot_name, slot_val = slot_info
                        predicted_slots[slot_name] = slot_val

        all_slots_correct = True

        for service_slot in teacher_action["service"]["slots"]:
            are_slots_equal = self.compare_slot_values(
                gt_slots, predicted_slots, service_slot
            )
            all_slots_correct = all_slots_correct and are_slots_equal
            if service_slot["name"] in predicted_slots:
                self.metrics.add("slot_p", AverageMetric(are_slots_equal))
            if service_slot["name"] in gt_slots:
                self.metrics.add("slot_r", AverageMetric(are_slots_equal))

        self.metrics.add("jga", AverageMetric(all_slots_correct))

    def setup_data(self, fold):
        schema_lookup, dialogues = self._load_data(fold)
        for dialogue in dialogues:
            turns = dialogue["turns"]
            context = []
            for turn_id in range(0, len(turns), 2):
                frames = turns[turn_id]["frames"]
                user_utterance = turns[turn_id]["utterance"]
                sys_utterance = turns[turn_id + 1]["utterance"]
                context.append(f"<user> {user_utterance} <system> {sys_utterance}")

                for frame in frames:
                    slot_values = frame["state"]["slot_values"]

                    label = self.SLOT_ENTRY_SEPARATOR.join(
                        [
                            f"{slot_name} {slot_val[0]}"
                            for slot_name, slot_val in slot_values.items()
                        ]
                    )

                    yield {
                        "text": " ".join(context),
                        "label": label,
                        "slots": slot_values,
                        "type": "text",
                        "service": schema_lookup[frame["service"]],
                    }, True


class SystemTeacher(GoogleSGDParser, tod_agents.TodSystemTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass


class UserSimulatorTeacher(GoogleSGDParser, tod_agents.TodUserSimulatorTeacher):
    pass


class StandaloneApiTeacher(GoogleSGDParser, tod_agents.TodStandaloneApiTeacher):
    pass


class SingleGoalAgent(GoogleSGDParser, tod_agents.TodSingleGoalAgent):
    pass


class GoalAgent(GoogleSGDParser, tod_agents.TodGoalAgent):
    pass


class ApiSchemaAgent(GoogleSGDParser, tod_agents.TodApiSchemaAgent):
    pass


class UserUttAgent(GoogleSGDParser, tod_agents.TodUserUttAgent):
    pass


class ApiCallAndSysUttAgent(GoogleSGDParser, tod_agents.TodApiCallAndSysUttAgent):
    pass



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
"""
GPT2 SeeKeR Agent for Language Modeling.
"""
from typing import Optional, Tuple, List, Union
import torch
import torch.nn

from parlai.agents.fid.fid import (
    FidModel,
    FidAgent,
    SearchQuerySearchEngineFiDAgent,
    WizIntGoldDocRetrieverFiDAgent,
)
from parlai.agents.hugging_face.gpt2 import Gpt2Agent
from parlai.agents.rag.rag import BaseGenerationAgentMixin
from parlai.agents.rag.retrievers import Document
from parlai.core.dict import DictionaryAgent
from parlai.core.message import Message
from parlai.core.opt import Opt
from parlai.core.params import ParlaiParser
from parlai.core.torch_agent import Batch, Output
from parlai.core.torch_generator_agent import TorchGeneratorAgent
from parlai.utils.torch import padded_tensor

from projects.seeker.agents.seeker import SeekerAgent
from projects.seeker.agents.gpt2_seeker_modules import (
    GPT2WithRetrieverModel,
    ComboGPT2Model,
)


class GPT2WithRetrieverAgentBase(Gpt2Agent, BaseGenerationAgentMixin):
    @staticmethod
    def build_rag_model(
        opt: Opt, dictionary: DictionaryAgent
    ) -> "GPT2WithRetrieverModel":
        return GPT2WithRetrieverModel(opt, dictionary)


class GPT2WithRetrieverAgent(FidAgent, GPT2WithRetrieverAgentBase):
    """
    GPT2 with Retriever agent.

    This agent packs in the retrieved documents and input context as one big "prompt" to
    the language model.
    """

    def __init__(self, opt, shared=None):
        opt['generation_model'] = 'gpt2'
        super().__init__(opt, shared)

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        FidAgent.add_cmdline_args(parser, partial_opt)
        Gpt2Agent.add_cmdline_args(parser, partial_opt)
        group = parser.add_argument_group('GPT2 Retriever Agent')
        group.add_argument(
            '--filter-docs-with-label',
            type='bool',
            default=False,
            help='If true, we make sure that we do not retrieve any docs containing '
            'label string VERBATIM.',
        )
        return parser

    @staticmethod
    def build_rag_model(
        opt: Opt, dictionary: DictionaryAgent
    ) -> "GPT2WithRetrieverModel":
        return GPT2WithRetrieverModel(opt, dictionary)

    def build_model(self) -> FidModel:
        return GPT2WithRetrieverModel(self.opt, self.dict)

    @property
    def generation_model(self) -> str:
        return self._generation_model

    @generation_model.setter
    def generation_model(self, model: str):
        """
        Override to always be GPT2.
        """
        self._generation_model = model
        self._generation_agent = GPT2WithRetrieverAgentBase

    def _pad_tensor(
        self, items: List[Union[List[int], torch.LongTensor]], is_label: bool = False
    ) -> Tuple[torch.LongTensor, List[int]]:
        """
        Override to always set fp16friendly to False and left_pad to True.
        """
        return padded_tensor(
            items, pad_idx=self.NULL_IDX, left_padded=True, fp16friendly=False
        )

    def _model_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        Optional[torch.LongTensor],
    ]:
        """
        Override model input to return target lengths as well.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.label_vec.ne(self.NULL_IDX).sum(1)
            if batch.label_vec is not None
            else None,
        )

    def _encoder_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        Optional[torch.LongTensor],
    ]:
        """
        During generation, we send the label truncation to the model.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.text_vec.new(batch.batchsize).fill_(self.label_truncate),
        )

    def eval_step(self, batch: Batch) -> Optional[Output]:
        """
        Override to cache the labels for retrieval.
        """
        if batch.label_vec is not None and self.opt.get('filter_docs_with_label'):
            self.model_api.set_labels(batch.label_vec)
        output = TorchGeneratorAgent.eval_step(self, batch)
        return output


class GPT2WithRetrieverSearchQueryAgent(
    GPT2WithRetrieverAgent, SearchQuerySearchEngineFiDAgent
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        GPT2WithRetrieverAgent.add_cmdline_args(parser, partial_opt)
        SearchQuerySearchEngineFiDAgent.add_cmdline_args(parser, partial_opt)
        return parser


class GPT2WithRetrieverGoldDocumentAgent(
    GPT2WithRetrieverAgent, WizIntGoldDocRetrieverFiDAgent
):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        GPT2WithRetrieverAgent.add_cmdline_args(parser, partial_opt)
        WizIntGoldDocRetrieverFiDAgent.add_cmdline_args(parser, partial_opt)
        return parser


####################
# Combo Agent Code #
####################
class GPT2ComboAgent(GPT2WithRetrieverAgent):
    """
    Combo GPT2 Agent.

    This agent can handle retrieval for some contexts, and vanilla decoder-only
    computation for others.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        """
        Add Combo Args.
        """
        super().add_cmdline_args(parser, partial_opt)
        gpt2_combo = parser.add_argument_group('GPT2 Combo Group')
        gpt2_combo.add_argument(
            '--skip-retrieval-key',
            type=str,
            default='skip_retrieval',
            help='key in observation determining whether to skip retrieval.',
        )

    def build_model(self) -> ComboGPT2Model:
        """
        Build and return ComboGPT2Model.
        """
        model = ComboGPT2Model(self.opt, self.dict)
        if self.opt['embedding_type'] != 'random':
            self._copy_embeddings(
                model.encoder.embeddings.weight, self.opt['embedding_type']
            )
        return model

    def batchify(self, obs_batch: List[Message], sort: bool = False) -> Batch:
        """
        Overrides batchify to add skip retrieval input vec.
        """
        batch = super().batchify(obs_batch, sort)
        valid_exs = [ex for ex in obs_batch if self.is_valid(ex)]
        if valid_exs:
            skip_retrieval = [
                ex.get(self.opt['skip_retrieval_key'], False) for ex in valid_exs
            ]
            batch.skip_retrieval_vec = torch.BoolTensor(skip_retrieval)
        return batch

    def _model_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
    ]:
        """
        Override _model_input to add skip_retrieval_vec.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.skip_retrieval_vec,
            batch.label_vec.ne(self.NULL_IDX).sum(1)
            if batch.label_vec is not None
            else None,
        )

    def _encoder_input(
        self, batch: Batch
    ) -> Tuple[
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.LongTensor,
        torch.BoolTensor,
        Optional[torch.LongTensor],
    ]:
        """
        For the encoder, during generation, we don't send label vec but rather the label
        truncation.
        """
        return (
            batch.text_vec,
            batch.text_vec.ne(self.NULL_IDX).sum(1),
            batch.query_vec,
            batch.input_turn_cnt_vec,
            batch.skip_retrieval_vec,
            batch.text_vec.new(batch.batchsize).fill_(self.label_truncate),
        )

    def get_retrieved_knowledge(self, message: Message) -> List[Document]:
        if message.get('skip_retrieval'):
            return []
        return super().get_retrieved_knowledge(message)

    def eval_step(self, batch: Batch) -> Optional[Output]:
        """
        Override to potentially filter docs with the label.

        Additionally add top docs to the output.
        """
        if batch.label_vec is not None and self.opt.get('filter_docs_with_label'):
            self.model_api.set_labels(batch.label_vec)
        output = TorchGeneratorAgent.eval_step(self, batch)
        if output is not None:
            output.top_docs = self.model_api.get_top_docs()
        return output


class GPT2ComboSearchQueryAgent(GPT2ComboAgent, SearchQuerySearchEngineFiDAgent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        GPT2ComboAgent.add_cmdline_args(parser, partial_opt)
        SearchQuerySearchEngineFiDAgent.add_cmdline_args(parser, partial_opt)
        return parser


class GPT2ComboGpt2GoldDocumentAgent(GPT2ComboAgent, WizIntGoldDocRetrieverFiDAgent):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        GPT2ComboAgent.add_cmdline_args(parser, partial_opt)
        WizIntGoldDocRetrieverFiDAgent.add_cmdline_args(parser, partial_opt)
        return parser


class GPT2SeekerAgent(SeekerAgent):
    @classmethod
    def get_additional_agent_args(cls) -> ParlaiParser:
        """
        Return a parser with arguments sourced from several sub models.
        """
        additional_agent_parser = SeekerAgent.get_additional_agent_args()
        GPT2ComboAgent.add_cmdline_args(additional_agent_parser)
        return additional_agent_parser



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.


from typing import Optional
from parlai.core.params import ParlaiParser
from parlai.core.opt import Opt
import os
import copy
from parlai.core.teachers import ParlAIDialogTeacher
from .build import build


def _path(opt):
    # Build the data if it doesn't exist.
    build(opt)
    dt = opt['datatype'].split(':')[0]
    return os.path.join(
        opt['datapath'], 'saferdialogues', 'saferdialogues_dataset', dt + '.txt'
    )


def _bad_path(opt):
    # Build the data if it doesn't exist.
    build(opt)
    dt = opt['datatype'].split(':')[0]
    if dt == 'valid' or dt == 'test':
        dt += '_bad'
    return os.path.join(
        opt['datapath'], 'saferdialogues', 'saferdialogues_dataset', dt + '.txt'
    )


class SaferDialoguesTeacher(ParlAIDialogTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('SaFeRDialogues options')
        agent.add_argument(
            '--recovery',
            type=bool,
            default=True,
            help="Whether or not to include the recovery utterance",
        )
        return parser

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['parlaidialogteacher_datafile'] = _path(opt)
        super().__init__(opt, shared)

    def _setup_data(self, path):
        super()._setup_data(path)
        if not self.opt['recovery']:
            for i, ep in enumerate(self.episodes):
                # make the signaling msg the label and remove the recovery msg
                texts = ep[0]['text'].split('\n')
                self.episodes[i][0].force_set('text', '\n'.join(texts[:-1]))
                self.episodes[i][0].force_set('labels', [texts[-1]])


class SaferDialoguesBADTeacher(ParlAIDialogTeacher):
    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        super().add_cmdline_args(parser, partial_opt)
        agent = parser.add_argument_group('SaFeRDialogues options')
        agent.add_argument(
            '--recovery',
            type=bool,
            default=True,
            help="Whether or not to include the recovery utterance",
        )
        return parser

    def __init__(self, opt, shared=None):
        opt = copy.deepcopy(opt)
        opt['parlaidialogteacher_datafile'] = _bad_path(opt)
        super().__init__(opt, shared)

    def _setup_data(self, path):
        super()._setup_data(path)
        if not self.opt['recovery']:
            for i, ep in enumerate(self.episodes):
                # make the signaling msg the label and remove the recovery msg
                texts = ep[0]['text'].split('\n')
                self.episodes[i][0].force_set('text', '\n'.join(texts[:-1]))
                self.episodes[i][0].force_set('labels', [texts[-1]])


class DefaultTeacher(SaferDialoguesTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .constants import LAYER_NORM_EPS  # noqa: F401
from .functions import (  # noqa: F401
    create_embeddings,
    create_position_codes,
    get_n_positions_from_options,
)
from .attention import BasicAttention, MultiHeadAttention  # noqa: F401
from .ffn import TransformerFFN  # noqa: F401
from .encoder import (  # noqa: F401
    PassThroughEncoder,
    TransformerEncoder,
    TransformerEncoderLayer,
)
from .decoder import (  # noqa: F401
    TransformerDecoder,
    TransformerDecoderLayer,
    TransformerDecoderOnly,
    TransformerDecoderOnlyLayer,
)
from .generator import TransformerGeneratorModel  # noqa: F401
from .wrappers import TransformerLinearWrapper, TransformerResponseWrapper  # noqa: F401
from .mem_net import TransformerMemNetModel  # noqa: F401



#!/usr/bin/env python3

# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Taskmaster-2 implementation for ParlAI.

No official train/valid/test splits are available as of 2020-05-18, so we make our own
splits.
"""

from parlai.core.params import ParlaiParser
import os
import pandas as pd
from collections import Counter
from parlai.core.opt import Opt
import parlai.core.tod.tod_core as tod
from parlai.utils.misc import warn_once
import json
from typing import Optional
from parlai.utils.data import DatatypeHelper
from parlai.utils.io import PathManager

import parlai.tasks.taskmaster2.build as build_
import parlai.core.tod.tod_agents as tod_agents


DOMAINS = [
    "flights",
    "food-ordering",
    "hotels",
    "movies",
    "restaurant-search",
    "sports",
    "music",
]


class Taskmaster2Parser(tod_agents.TodStructuredDataParser):
    """
    Abstract data loader.
    """

    @classmethod
    def add_cmdline_args(
        cls, parser: ParlaiParser, partial_opt: Optional[Opt] = None
    ) -> ParlaiParser:
        parser.add_argument(
            "--taskmaster2-domains",
            nargs="+",
            default=DOMAINS,
            choices=DOMAINS,
            help="Uses last passed in configuration.",
        )
        parser.add_argument(
            "--use-cumulative-api-calls",
            type=bool,
            default=True,
            help="Have API Call/API response turns only when an API response"
            "slot exist. Accumulate all API call slots with same API call name",
        )
        return super().add_cmdline_args(parser, partial_opt)

    def __init__(self, opt: Opt, shared=None):
        self.fold = DatatypeHelper.fold(opt["datatype"])
        opt["datafile"] = self.fold
        self.dpath = os.path.join(opt["datapath"], "taskmaster-2")
        if shared is None:
            warn_once(
                "Taskmaster2 is a beta dataset, and format may significantly change."
            )
            build_.build(opt)
        super().__init__(opt, shared)

    def _load_data(self, fold, domains):
        # load up the ontology
        ontologies = {}
        for section in domains:
            fn = os.path.join(self.dpath, section + ".onto.json")
            with PathManager.open(fn, "r") as f:
                ontologies.update(json.load(f))

        chunks = []
        for section in domains:
            with PathManager.open(os.path.join(self.dpath, section + ".json")) as f:
                subset = pd.read_json(f)
            subset["domain"] = section
            chunks.append(subset)
        chunks = pd.concat(chunks, axis=0)
        # deterministic shuffle data for splits
        chunks = chunks.sample(frac=1.0, random_state=42)
        split_size = len(chunks) // 10
        if fold == "train":
            chunks = chunks[: split_size * 8]
        elif fold == "valid":
            chunks = chunks[split_size * 8 : split_size * 9]
        elif fold == "test":
            chunks = chunks[split_size * 9 :]
        return chunks, ontologies

    def _parse_segment_to_slots(self, segment_list):
        result = {}
        for segment in segment_list:
            slot_name = segment["annotations"][0]["name"]
            slot_value = segment["text"]
            prefix_split_idx = slot_name.find(".")
            api_name = slot_name[:prefix_split_idx]
            slot_name = slot_name[prefix_split_idx + 1 :]
            result[slot_name] = slot_value
            result[tod.STANDARD_API_NAME_SLOT] = api_name
        return result

    def _get_utterance_and_api_call_for_speaker(self, speaker, utterances, idx):
        utts = []
        slots = {}
        while idx < len(utterances):
            here = utterances[idx]
            if here["speaker"] != speaker:
                break
            utts.append(here["text"])
            slots.update(self._parse_segment_to_slots(here.get("segments", [])))
            idx += 1
        return idx, "\n".join(utts), slots

    def _get_onto_list(self, onto_map, domain):
        results = []
        domain = domain.replace(
            "-", "_"
        )  # cause they changed it for restaurant-search >.>
        for data in onto_map[domain]:
            call = {}
            call[tod.STANDARD_API_NAME_SLOT] = data["prefix"]
            call[tod.STANDARD_OPTIONAL_KEY] = data[
                "annotations"
            ]  # make all args optional since not specified
            results.append(call)
        return results

    def setup_episodes(self, fold):
        """
        Parses into TodStructuredEpisode.
        """
        domains = self.opt.get("taskmaster2_domains", DOMAINS)
        chunks, ontologies = self._load_data(fold, domains)
        domains_cnt = Counter()
        episodes = []
        for _, row in chunks.iterrows():
            domains_cnt[row["domain"]] += 1
            utterances = row["utterances"][:]

            idx = 0
            rounds = []
            goal_calls = []
            if len(utterances) > 0 and utterances[0]["speaker"] == "ASSISTANT":
                idx, sys_utt, api_resp = self._get_utterance_and_api_call_for_speaker(
                    "ASSISTANT", utterances, idx
                )
                r = tod.TodStructuredRound(api_resp_machine=api_resp, sys_utt=sys_utt)
                rounds.append(r)

            cum_api_call = {}
            while idx < len(utterances):
                idx, user_utt, api_call = self._get_utterance_and_api_call_for_speaker(
                    "USER", utterances, idx
                )
                idx, sys_utt, api_resp = self._get_utterance_and_api_call_for_speaker(
                    "ASSISTANT", utterances, idx
                )
                if not self.opt["use_cumulative_api_calls"]:
                    r = tod.TodStructuredRound(
                        user_utt=user_utt,
                        api_call_machine=api_call,
                        api_resp_machine=api_resp,
                        sys_utt=sys_utt,
                    )
                else:
                    cum_api_call = self.process_call_for_cumlative_standalone_api(
                        api_call, cum_api_call
                    )
                    r = tod.TodStructuredRound(
                        user_utt=user_utt,
                        api_call_machine=cum_api_call if len(api_resp) > 0 else {},
                        api_resp_machine=api_resp if len(api_resp) > 0 else {},
                        sys_utt=sys_utt,
                    )

                rounds.append(r)
                if len(api_call) > 0:
                    goal_calls.append(api_call)

            episode = tod.TodStructuredEpisode(
                domain=tod.SerializationHelpers.inner_list_join(row["domain"]),
                api_schemas_machine=self._get_onto_list(ontologies, row["domain"]),
                goal_calls_machine=goal_calls,
                rounds=rounds,
                delex=self.opt.get("delex", False),
            )
            episodes.append(episode)
        return episodes

    def get_id_task_prefix(self):
        return "Taskmaster2"

    def _label_fold(self, chunks):
        return chunks.conversation_id.apply(self._h)

    def process_call_for_cumlative_standalone_api(self, new_call, cum_calls):
        if (
            len(new_call) > 0
            and len(cum_calls) > 0
            and new_call[tod.STANDARD_API_NAME_SLOT]
            != cum_calls[tod.STANDARD_API_NAME_SLOT]
        ):
            cum_calls = {}
        cum_calls.update(new_call)
        return cum_calls


class UserSimulatorTeacher(Taskmaster2Parser, tod_agents.TodUserSimulatorTeacher):
    pass


class SystemTeacher(Taskmaster2Parser, tod_agents.TodSystemTeacher):
    pass


class DefaultTeacher(SystemTeacher):
    pass



#!/usr/bin/env python3
# Copyright (c) Facebook, Inc. and its affiliates.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
from parlai.tasks.msc.agents import PersonaSummaryTeacher
import parlai.utils.logging as logging
from projects.bb3.tasks.module_level_tasks import BB3TeacherMixin


class MscMemoryGenerationTeacher(BB3TeacherMixin, PersonaSummaryTeacher):
    def __init__(self, opt, shared=None):
        opt['include_last_session'] = True
        opt['nopersona_subsampling_weight'] = 0.05
        mutators = '+'.join(['skip_retrieval_mutator', 'prompt_memory_mutator'])
        if opt.get('mutators'):
            mutators = '+'.join([mutators, opt['mutators']])
        logging.warning(f'overriding mutators to {mutators}')
        opt['mutators'] = mutators
        super().__init__(opt, shared)
